from descope.common import (
    COOKIE_DATA_NAME,
    REFRESH_SESSION_COOKIE_NAME,
    REFRESH_SESSION_TOKEN_NAME,
    SESSION_COOKIE_NAME,
    SESSION_TOKEN_NAME,
    AccessKeyLoginOptions,
    DeliveryMethod,
    LoginOptions,
    SignUpOptions,
)
from descope.descope_client import DescopeClient
from descope.exceptions import (
    API_RATE_LIMIT_RETRY_AFTER_HEADER,
    ERROR_TYPE_API_RATE_LIMIT,
    ERROR_TYPE_SERVER_ERROR,
    AuthException,
    RateLimitException,
)
from descope.management.common import (
    AssociatedTenant,
    SAMLIDPAttributeMappingInfo,
    SAMLIDPGroupsMappingInfo,
    SAMLIDPRoleGroupMappingInfo,
)
from descope.management.sso_settings import (
    AttributeMapping,
    OIDCAttributeMapping,
    RoleMapping,
    SSOOIDCSettings,
    SSOSAMLSettings,
    SSOSAMLSettingsByMetadata,
)
from descope.management.user import UserObj
from descope.management.user_pwd import (
    UserPassword,
    UserPasswordBcrypt,
    UserPasswordDjango,
    UserPasswordFirebase,
    UserPasswordPbkdf2,
)



from __future__ import annotations

import requests

from descope._auth_base import AuthBase
from descope.auth import Auth
from descope.common import (
    REFRESH_SESSION_COOKIE_NAME,
    DeliveryMethod,
    EndpointsV1,
    LoginOptions,
    SignUpOptions,
    signup_options_to_dict,
    validate_refresh_token_provided,
)
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException


class EnchantedLink(AuthBase):
    def sign_in(
        self,
        login_id: str,
        uri: str,
        login_options: LoginOptions | None = None,
        refresh_token: str | None = None,
    ) -> dict:
        if not login_id:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "login_id is empty",
            )

        validate_refresh_token_provided(login_options, refresh_token)

        body = EnchantedLink._compose_signin_body(login_id, uri, login_options)
        url = EnchantedLink._compose_signin_url()
        response = self._http.post(url, body=body, pswd=refresh_token)
        return EnchantedLink._get_pending_ref_from_response(response)

    def sign_up(
        self,
        login_id: str,
        uri: str,
        user: dict | None,
        signup_options: SignUpOptions | None = None,
    ) -> dict:
        if not user:
            user = {}

        if not self._auth.adjust_and_verify_delivery_method(
            DeliveryMethod.EMAIL, login_id, user
        ):
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"Login ID {login_id} is not valid for email",
            )

        body = EnchantedLink._compose_signup_body(login_id, uri, user, signup_options)
        url = EnchantedLink._compose_signup_url()
        response = self._http.post(url, body=body)
        return EnchantedLink._get_pending_ref_from_response(response)

    def sign_up_or_in(
        self, login_id: str, uri: str, signup_options: SignUpOptions | None = None
    ) -> dict:
        login_options: LoginOptions | None = None
        if signup_options is not None:
            login_options = LoginOptions(
                custom_claims=signup_options.customClaims,
                template_options=signup_options.templateOptions,
                template_id=signup_options.templateId,
            )

        body = EnchantedLink._compose_signin_body(
            login_id,
            uri,
            login_options,
        )
        url = EnchantedLink._compose_sign_up_or_in_url()
        response = self._http.post(url, body=body)
        return EnchantedLink._get_pending_ref_from_response(response)

    def get_session(self, pending_ref: str) -> dict:
        uri = EndpointsV1.get_session_enchantedlink_auth_path
        body = EnchantedLink._compose_get_session_body(pending_ref)
        response = self._http.post(uri, body=body)
        resp = response.json()
        jwt_response = self._auth.generate_jwt_response(
            resp, response.cookies.get(REFRESH_SESSION_COOKIE_NAME, None), None
        )
        return jwt_response

    def verify(self, token: str):
        uri = EndpointsV1.verify_enchantedlink_auth_path
        body = EnchantedLink._compose_verify_body(token)
        self._http.post(uri, body=body)

    def update_user_email(
        self,
        login_id: str,
        email: str,
        refresh_token: str,
        add_to_login_ids: bool = False,
        on_merge_use_existing: bool = False,
        template_options: dict | None = None,
        template_id: str | None = None,
        provider_id: str | None = None,
    ) -> dict:
        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        Auth.validate_email(email)

        body = EnchantedLink._compose_update_user_email_body(
            login_id,
            email,
            add_to_login_ids,
            on_merge_use_existing,
            template_options,
            template_id,
            provider_id,
        )
        uri = EndpointsV1.update_user_email_enchantedlink_path
        response = self._http.post(uri, body=body, pswd=refresh_token)
        return EnchantedLink._get_pending_ref_from_response(response)

    @staticmethod
    def _compose_signin_url() -> str:
        return Auth.compose_url(
            EndpointsV1.sign_in_auth_enchantedlink_path, DeliveryMethod.EMAIL
        )

    @staticmethod
    def _compose_signup_url() -> str:
        return Auth.compose_url(
            EndpointsV1.sign_up_auth_enchantedlink_path, DeliveryMethod.EMAIL
        )

    @staticmethod
    def _compose_sign_up_or_in_url() -> str:
        return Auth.compose_url(
            EndpointsV1.sign_up_or_in_auth_enchantedlink_path, DeliveryMethod.EMAIL
        )

    @staticmethod
    def _compose_signin_body(
        login_id: str,
        uri: str,
        login_options: LoginOptions | None = None,
    ) -> dict:
        return {
            "loginId": login_id,
            "URI": uri,
            "loginOptions": login_options.__dict__ if login_options else {},
        }

    @staticmethod
    def _compose_signup_body(
        login_id: str,
        uri: str,
        user: dict | None = None,
        signup_options: SignUpOptions | None = None,
    ) -> dict:
        body: dict[str, str | bool | dict] = {"loginId": login_id, "URI": uri}

        if signup_options is not None:
            body["loginOptions"] = signup_options_to_dict(signup_options)

        if user is not None:
            body["user"] = user
            method_str, val = Auth.get_login_id_by_method(DeliveryMethod.EMAIL, user)
            body[method_str] = val
        return body

    @staticmethod
    def _compose_verify_body(token: str) -> dict:
        return {"token": token}

    @staticmethod
    def _compose_update_user_email_body(
        login_id: str,
        email: str,
        add_to_login_ids: bool,
        on_merge_use_existing: bool,
        template_options: dict | None = None,
        template_id: str | None = None,
        provider_id: str | None = None,
    ) -> dict:
        body: dict[str, str | bool | dict] = {
            "loginId": login_id,
            "email": email,
            "addToLoginIDs": add_to_login_ids,
            "onMergeUseExisting": on_merge_use_existing,
        }
        if template_options is not None:
            body["templateOptions"] = template_options
        if template_id is not None:
            body["templateId"] = template_id
        if provider_id is not None:
            body["providerId"] = provider_id

        return body

    @staticmethod
    def _compose_get_session_body(pending_ref: str) -> dict:
        return {"pendingRef": pending_ref}

    @staticmethod
    def _get_pending_ref_from_response(response: requests.Response) -> dict:
        return response.json()



from typing import Optional

from descope.auth import Auth
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException
from descope.http_client import HTTPClient
from descope.management.access_key import AccessKey
from descope.management.audit import Audit
from descope.management.authz import Authz
from descope.management.fga import FGA
from descope.management.flow import Flow
from descope.management.group import Group
from descope.management.jwt import JWT
from descope.management.outbound_application import (
    OutboundApplication,
    OutboundApplicationByToken,
)
from descope.management.permission import Permission
from descope.management.project import Project
from descope.management.role import Role
from descope.management.sso_application import SSOApplication
from descope.management.sso_settings import SSOSettings

# Import management modules after adapter to avoid circularities
from descope.management.tenant import Tenant
from descope.management.user import User


class MGMT:
    _http: HTTPClient

    def __init__(
        self, http_client: HTTPClient, auth: Auth, fga_cache_url: Optional[str] = None
    ):
        """Create a management API facade.

        Args:
            http_client: HTTP client to use for all management HTTP calls.
        """
        self._http = http_client
        self._access_key = AccessKey(http_client)
        self._audit = Audit(http_client)
        self._authz = Authz(http_client)
        self._fga = FGA(http_client, fga_cache_url=fga_cache_url)
        self._flow = Flow(http_client)
        self._group = Group(http_client)
        self._jwt = JWT(http_client, auth=auth)
        self._outbound_application = OutboundApplication(http_client)
        self._outbound_application_by_token = OutboundApplicationByToken(http_client)
        self._permission = Permission(http_client)
        self._project = Project(http_client)
        self._role = Role(http_client)
        self._sso = SSOSettings(http_client)
        self._sso_application = SSOApplication(http_client)
        self._tenant = Tenant(http_client)
        self._user = User(http_client)

    def _ensure_management_key(self, property_name: str):
        """Check if management key is available for the given property."""
        if not self._http.management_key:
            raise AuthException(
                error_type=ERROR_TYPE_INVALID_ARGUMENT,
                error_message=f"Management key is required to access '{property_name}' functionality",
            )

    @property
    def tenant(self):
        self._ensure_management_key("tenant")
        return self._tenant

    @property
    def sso_application(self):
        self._ensure_management_key("sso_application")
        return self._sso_application

    @property
    def user(self):
        self._ensure_management_key("user")
        return self._user

    @property
    def access_key(self):
        self._ensure_management_key("access_key")
        return self._access_key

    @property
    def sso(self):
        self._ensure_management_key("sso")
        return self._sso

    @property
    def jwt(self):
        self._ensure_management_key("jwt")
        return self._jwt

    @property
    def permission(self):
        self._ensure_management_key("permission")
        return self._permission

    @property
    def role(self):
        self._ensure_management_key("role")
        return self._role

    @property
    def group(self):
        self._ensure_management_key("group")
        return self._group

    @property
    def flow(self):
        self._ensure_management_key("flow")
        return self._flow

    @property
    def audit(self):
        self._ensure_management_key("audit")
        return self._audit

    @property
    def authz(self):
        self._ensure_management_key("authz")
        return self._authz

    @property
    def fga(self):
        self._ensure_management_key("fga")
        return self._fga

    @property
    def project(self):
        self._ensure_management_key("project")
        return self._project

    @property
    def outbound_application(self):
        self._ensure_management_key("outbound_application")
        return self._outbound_application

    @property
    def outbound_application_by_token(self):
        # No management key check for outbound_app_token (as authentication for those methods is done by inbound app token)
        return self._outbound_application_by_token



from typing import Optional

from descope._http_base import HTTPBase
from descope.auth import Auth
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException
from descope.jwt_common import generate_jwt_response
from descope.management.common import (
    MgmtLoginOptions,
    MgmtSignUpOptions,
    MgmtUserRequest,
    MgmtV1,
    is_jwt_required,
)


class JWT(HTTPBase):
    _auth: Auth

    def __init__(self, http_client, auth: Auth):
        super().__init__(http_client)
        self._auth = auth

    def update_jwt(
        self, jwt: str, custom_claims: dict, refresh_duration: int = 0
    ) -> str:
        """
        Given a valid JWT, update it with custom claims, and update its authz claims as well

        Args:
        token (str): valid jwt.
        custom_claims (dict): Custom claims to add to JWT, system claims will be filtered out
        refresh_duration (int): duration in seconds for which the new JWT will be valid

        Return value (str): the newly updated JWT

        Raise:
        AuthException: raised if update failed
        """
        if not jwt:
            raise AuthException(400, ERROR_TYPE_INVALID_ARGUMENT, "jwt cannot be empty")
        response = self._http.post(
            MgmtV1.update_jwt_path,
            body={
                "jwt": jwt,
                "customClaims": custom_claims,
                "refreshDuration": refresh_duration,
            },
            params=None,
        )
        return response.json().get("jwt", "")

    def impersonate(
        self,
        impersonator_id: str,
        login_id: str,
        validate_consent: bool,
        custom_claims: Optional[dict] = None,
        tenant_id: Optional[str] = None,
        refresh_duration: Optional[int] = None,
    ) -> str:
        """
        Impersonate to another user

        Args:
        impersonator_id (str): login id / user id of impersonator, must have "impersonation" permission.
        login_id (str): login id of the user whom to which to impersonate to.
        validate_consent (bool): Indicate whether to allow impersonation in any case or only if a consent to this operation was granted.
        customClaims dict: Custom claims to add to JWT
        tenant_id (str): tenant id to set on DCT claim.
        refresh_duration (int): duration in seconds for which the new JWT will be valid

        Return value (str): A JWT of the impersonated user

        Raise:
        AuthException: raised if update failed
        """
        if not impersonator_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "impersonator_id cannot be empty"
            )
        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "login_id cannot be empty"
            )
        response = self._http.post(
            MgmtV1.impersonate_path,
            body={
                "loginId": login_id,
                "impersonatorId": impersonator_id,
                "validateConsent": validate_consent,
                "customClaims": custom_claims,
                "selectedTenant": tenant_id,
                "refreshDuration": refresh_duration,
            },
            params=None,
        )
        return response.json().get("jwt", "")

    def stop_impersonation(
        self,
        jwt: str,
        custom_claims: Optional[dict] = None,
        tenant_id: Optional[str] = None,
        refresh_duration: Optional[int] = None,
    ) -> str:
        """
        Stop impersonation and return to the original user
        Args:
        jwt (str): The impersonation jwt to stop.
        customClaims dict: Custom claims to add to JWT
        tenant_id (str): tenant id to set on DCT claim.
        refresh_duration (int): duration in seconds for which the new JWT will be valid

        Return value (str): A JWT of the actor

        Raise:
        AuthException: raised if update failed
        """
        if not jwt or jwt == "":
            raise AuthException(400, ERROR_TYPE_INVALID_ARGUMENT, "jwt cannot be empty")

        response = self._http.post(
            MgmtV1.stop_impersonation_path,
            body={
                "jwt": jwt,
                "customClaims": custom_claims,
                "selectedTenant": tenant_id,
                "refreshDuration": refresh_duration,
            },
            params=None,
        )
        return response.json().get("jwt", "")

    def sign_in(
        self, login_id: str, login_options: Optional[MgmtLoginOptions] = None
    ) -> dict:
        """
        Generate a JWT for a user, simulating a signin request.

        Args:
        login_id (str): login id of the user.
        login_options (MgmtLoginOptions): options for the login request.
        """

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "login_id cannot be empty"
            )

        if login_options is None:
            login_options = MgmtLoginOptions()

        if is_jwt_required(login_options) and not login_options.jwt:
            raise AuthException(400, ERROR_TYPE_INVALID_ARGUMENT, "JWT is required")

        response = self._http.post(
            MgmtV1.mgmt_sign_in_path,
            body={
                "loginId": login_id,
                "stepup": login_options.stepup,
                "mfa": login_options.mfa,
                "revokeOtherSessions": login_options.revoke_other_sessions,
                "customClaims": login_options.custom_claims,
                "jwt": login_options.jwt,
                "refreshDuration": login_options.refresh_duration,
            },
            params=None,
        )
        resp = response.json()
        jwt_response = generate_jwt_response(
            resp, None, None, self._auth.validate_token
        )
        return jwt_response

    def sign_up(
        self,
        login_id: str,
        user: Optional[MgmtUserRequest] = None,
        signup_options: Optional[MgmtSignUpOptions] = None,
    ) -> dict:
        """
        Generate a JWT for a user, simulating a signup request.

        Args:
        login_id (str): login id of the user.
        user (MgmtUserRequest): user details.
        signup_options (MgmtSignUpOptions): signup options.
        """

        return self._sign_up_internal(
            login_id, MgmtV1.mgmt_sign_up_path, user, signup_options
        )

    def sign_up_or_in(
        self,
        login_id: str,
        user: Optional[MgmtUserRequest] = None,
        signup_options: Optional[MgmtSignUpOptions] = None,
    ) -> dict:
        """
        Generate a JWT for a user, simulating a signup or in request.

        Args:
        login_id (str): login id of the user.
        user (MgmtUserRequest): user details.
        signup_options (MgmtSignUpOptions): signup options.
        """
        return self._sign_up_internal(
            login_id, MgmtV1.mgmt_sign_up_or_in_path, user, signup_options
        )

    def _sign_up_internal(
        self,
        login_id: str,
        endpoint: str,
        user: Optional[MgmtUserRequest] = None,
        signup_options: Optional[MgmtSignUpOptions] = None,
    ) -> dict:
        if user is None:
            user = MgmtUserRequest()

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "login_id cannot be empty"
            )

        if signup_options is None:
            signup_options = MgmtSignUpOptions()

        response = self._http.post(
            endpoint,
            body={
                "loginId": login_id,
                "user": user.to_dict(),
                "emailVerified": user.email_verified,
                "phoneVerified": user.phone_verified,
                "ssoAppId": user.sso_app_id,
                "customClaims": signup_options.custom_claims,
                "refreshDuration": signup_options.refresh_duration,
            },
            params=None,
        )
        resp = response.json()
        jwt_response = generate_jwt_response(
            resp, None, None, self._auth.validate_token
        )
        return jwt_response

    def anonymous(
        self,
        custom_claims: Optional[dict] = None,
        tenant_id: Optional[str] = None,
        refresh_duration: Optional[int] = None,
    ) -> dict:
        """
        Generate a JWT for an anonymous user.

        Args:
        custom_claims dict: Custom claims to add to JWT
        tenant_id (str): tenant id to set on DCT claim.
        """

        response = self._http.post(
            MgmtV1.anonymous_path,
            body={
                "customClaims": custom_claims,
                "selectedTenant": tenant_id,
                "refreshDuration": refresh_duration,
            },
            params=None,
        )
        resp = response.json()
        jwt_response = generate_jwt_response(
            resp, None, None, self._auth.validate_token
        )
        del jwt_response["firstSeen"]
        del jwt_response["user"]
        return jwt_response



from typing import Any, List, Optional, Union

from descope._http_base import HTTPBase
from descope.common import DeliveryMethod, LoginOptions, get_method_string
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException
from descope.management.common import (
    AssociatedTenant,
    MgmtV1,
    Sort,
    associated_tenants_to_dict,
    sort_to_dict,
)
from descope.management.user_pwd import UserPassword


class UserObj:
    def __init__(
        self,
        login_id: str,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        role_names: Optional[List[str]] = None,
        user_tenants: Optional[List[AssociatedTenant]] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        additional_login_ids: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
        password: Optional[UserPassword] = None,
        seed: Optional[str] = None,
        status: Optional[str] = None,
    ):
        self.login_id = login_id
        self.email = email
        self.phone = phone
        self.display_name = display_name
        self.given_name = given_name
        self.middle_name = middle_name
        self.family_name = family_name
        self.role_names = role_names
        self.user_tenants = user_tenants
        self.picture = picture
        self.custom_attributes = custom_attributes
        self.verified_email = verified_email
        self.verified_phone = verified_phone
        self.additional_login_ids = additional_login_ids
        self.sso_app_ids = sso_app_ids
        self.password = password
        self.seed = seed
        self.status = status


class CreateUserObj:
    def __init__(
        self,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
    ):
        self.email = email
        self.phone = phone
        self.name = name
        self.given_name = given_name
        self.middle_name = middle_name
        self.family_name = family_name


class User(HTTPBase):
    def create(
        self,
        login_id: str,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        role_names: Optional[List[str]] = None,
        user_tenants: Optional[List[AssociatedTenant]] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        invite_url: Optional[str] = None,
        additional_login_ids: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
    ) -> dict:
        """
        Create a new user. Users can have any number of optional fields, including email, phone number and authorization.

        Args:
        login_id (str): user login ID.
        email (str): Optional user email address.
        phone (str): Optional user phone number.
        display_name (str): Optional user display name.
        role_names (List[str]): An optional list of the user's roles without tenant association. These roles are
            mutually exclusive with the `user_tenant` roles.
        user_tenants (List[AssociatedTenant]): An optional list of the user's tenants, and optionally, their roles per tenant. These roles are
            mutually exclusive with the general `role_names`.
        picture (str): Optional url for user picture
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        sso_app_ids (List[str]): Optional, list of SSO applications IDs to be associated with the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the created user information.

        Raise:
        AuthException: raised if create operation fails
        """
        role_names = [] if role_names is None else role_names
        user_tenants = [] if user_tenants is None else user_tenants

        response = self._http.post(
            MgmtV1.user_create_path,
            body=User._compose_create_body(
                login_id,
                email,
                phone,
                display_name,
                given_name,
                middle_name,
                family_name,
                role_names,
                user_tenants,
                False,
                False,
                picture,
                custom_attributes,
                verified_email,
                verified_phone,
                invite_url,
                None,
                None,
                additional_login_ids,
                sso_app_ids,
            ),
        )
        return response.json()

    def create_test_user(
        self,
        login_id: str,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        role_names: Optional[List[str]] = None,
        user_tenants: Optional[List[AssociatedTenant]] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        invite_url: Optional[str] = None,
        additional_login_ids: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
    ) -> dict:
        """
        Create a new test user.
        The login_id is required and will determine what the user will use to sign in.
        Make sure the login id is unique for test. All other fields are optional.

        Args:
        login_id (str): user login ID.
        email (str): Optional user email address.
        phone (str): Optional user phone number.
        display_name (str): Optional user display name.
        role_names (List[str]): An optional list of the user's roles without tenant association. These roles are
            mutually exclusive with the `user_tenant` roles.
        user_tenants (List[AssociatedTenant]): An optional list of the user's tenants, and optionally, their roles per tenant. These roles are
            mutually exclusive with the general `role_names`.
        picture (str): Optional url for user picture
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        sso_app_ids (List[str]): Optional, list of SSO applications IDs to be associated with the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the created test user information.

        Raise:
        AuthException: raised if create operation fails
        """
        role_names = [] if role_names is None else role_names
        user_tenants = [] if user_tenants is None else user_tenants

        response = self._http.post(
            MgmtV1.test_user_create_path,
            body=User._compose_create_body(
                login_id,
                email,
                phone,
                display_name,
                given_name,
                middle_name,
                family_name,
                role_names,
                user_tenants,
                False,
                True,
                picture,
                custom_attributes,
                verified_email,
                verified_phone,
                invite_url,
                None,
                None,
                additional_login_ids,
                sso_app_ids,
            ),
        )
        return response.json()

    def invite(
        self,
        login_id: str,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        role_names: Optional[List[str]] = None,
        user_tenants: Optional[List[AssociatedTenant]] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        invite_url: Optional[str] = None,
        send_mail: Optional[
            bool
        ] = None,  # send invite via mail, default is according to project settings
        send_sms: Optional[
            bool
        ] = None,  # send invite via text message, default is according to project settings
        additional_login_ids: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
        template_id: str = "",
        test: bool = False,
    ) -> dict:
        """
        Create a new user and invite them via an email / text message.

        Functions exactly the same as the `create` function with the additional invitation
            behavior. See the documentation above for the general creation behavior.

        IMPORTANT: Since the invitation is sent by email / phone, make sure either
            the email / phone is explicitly set, or the login_id itself is an email address / phone number.
            You must configure the invitation URL in the Descope console prior to
            calling the method.
        """
        role_names = [] if role_names is None else role_names
        user_tenants = [] if user_tenants is None else user_tenants

        response = self._http.post(
            MgmtV1.user_create_path,
            body=User._compose_create_body(
                login_id,
                email,
                phone,
                display_name,
                given_name,
                middle_name,
                family_name,
                role_names,
                user_tenants,
                True,
                test,
                picture,
                custom_attributes,
                verified_email,
                verified_phone,
                invite_url,
                send_mail,
                send_sms,
                additional_login_ids,
                sso_app_ids,
                template_id,
            ),
        )
        return response.json()

    def invite_batch(
        self,
        users: List[UserObj],
        invite_url: Optional[str] = None,
        send_mail: Optional[
            bool
        ] = None,  # send invite via mail, default is according to project settings
        send_sms: Optional[
            bool
        ] = None,  # send invite via text message, default is according to project settings
    ) -> dict:
        """
        Create users in batch and invite them via an email / text message.

        Functions exactly the same as the `create` function with the additional invitation
            behavior. See the documentation above for the general creation behavior.

        IMPORTANT: Since the invitation is sent by email / phone, make sure either
            the email / phone is explicitly set, or the login_id itself is an email address / phone number.
            You must configure the invitation URL in the Descope console prior to
            calling the method.
        """

        response = self._http.post(
            MgmtV1.user_create_batch_path,
            body=User._compose_create_batch_body(
                users,
                invite_url,
                send_mail,
                send_sms,
            ),
        )
        return response.json()

    def update(
        self,
        login_id: str,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        role_names: Optional[List[str]] = None,
        user_tenants: Optional[List[AssociatedTenant]] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        additional_login_ids: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
        test: bool = False,
    ) -> dict:
        """
        Update an existing user with the given various fields. IMPORTANT: All parameters are used as overrides
        to the existing user. Empty fields will override populated fields. Use carefully.
        Use `patch` for partial updates instead.

        Args:
        login_id (str): The login ID of the user to update.
        email (str): Optional user email address.
        phone (str): Optional user phone number.
        display_name (str): Optional user display name.
        given_name (str): Optional user given name.
        middle_name (str): Optional user middle name.
        family_name (str): Optional user family name.
        role_names (List[str]): An optional list of the user's roles without tenant association. These roles are
            mutually exclusive with the `user_tenant` roles.
        user_tenants (List[AssociatedTenant]): An optional list of the user's tenants, and optionally, their roles per tenant. These roles are
            mutually exclusive with the general `role_names`.
        picture (str): Optional url for user picture
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        sso_app_ids (List[str]): Optional, list of SSO applications IDs to be associated with the user.
        test (bool, optional): Set to True to update a test user. Defaults to False.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if update operation fails
        """
        role_names = [] if role_names is None else role_names
        user_tenants = [] if user_tenants is None else user_tenants

        response = self._http.post(
            MgmtV1.user_update_path,
            body=User._compose_update_body(
                login_id,
                email,
                phone,
                display_name,
                given_name,
                middle_name,
                family_name,
                role_names,
                user_tenants,
                test,
                picture,
                custom_attributes,
                verified_email,
                verified_phone,
                additional_login_ids,
                sso_app_ids,
                None,
            ),
        )
        return response.json()

    def patch(
        self,
        login_id: str,
        email: Optional[str] = None,
        phone: Optional[str] = None,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        role_names: Optional[List[str]] = None,
        user_tenants: Optional[List[AssociatedTenant]] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        sso_app_ids: Optional[List[str]] = None,
        status: Optional[str] = None,
        test: bool = False,
    ) -> dict:
        """
        Patches an existing user with the given various fields. Only the given fields will be used to update the user.

        Args:
        login_id (str): The login ID of the user to update.
        email (str): Optional user email address.
        phone (str): Optional user phone number.
        display_name (str): Optional user display name.
        given_name (str): Optional user given name.
        middle_name (str): Optional user middle name.
        family_name (str): Optional user family name.
        role_names (List[str]): An optional list of the user's roles without tenant association. These roles are
            mutually exclusive with the `user_tenant` roles.
        user_tenants (List[AssociatedTenant]): An optional list of the user's tenants, and optionally, their roles per tenant. These roles are
            mutually exclusive with the general `role_names`.
        picture (str): Optional url for user picture
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        sso_app_ids (List[str]): Optional, list of SSO applications IDs to be associated with the user.
        status (str): Optional status field. Can be one of: "enabled", "disabled", "invited".
        test (bool, optional): Set to True to update a test user. Defaults to False.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the patched user information.

        Raise:
        AuthException: raised if patch operation fails
        """
        if status is not None and status not in ["enabled", "disabled", "invited"]:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"Invalid status value: {status}. Must be one of: enabled, disabled, invited",
            )
        response = self._http.patch(
            MgmtV1.user_patch_path,
            body=User._compose_patch_body(
                login_id,
                email,
                phone,
                display_name,
                given_name,
                middle_name,
                family_name,
                role_names,
                user_tenants,
                picture,
                custom_attributes,
                verified_email,
                verified_phone,
                sso_app_ids,
                status,
                test,
            ),
        )
        return response.json()

    def patch_batch(
        self,
        users: List[UserObj],
        test: bool = False,
    ) -> dict:
        """
        Patch users in batch. Only the provided fields will be updated for each user.

        Args:
        users (List[UserObj]): A list of UserObj instances representing users to be patched.
            Each UserObj should have a login_id and the fields to be updated.
        test (bool, optional): Set to True to patch test users. Defaults to False.

        Return value (dict):
        Return dict in the format
             {"patchedUsers": [...], "failedUsers": [...]}
        "patchedUsers" contains successfully patched users,
        "failedUsers" contains users that failed to be patched with error details.

        Raise:
        AuthException: raised if patch batch operation fails
        """
        # Validate status fields for all users
        for user in users:
            if user.status is not None and user.status not in [
                "enabled",
                "disabled",
                "invited",
            ]:
                raise AuthException(
                    400,
                    ERROR_TYPE_INVALID_ARGUMENT,
                    f"Invalid status value: {user.status} for user {user.login_id}. Must be one of: enabled, disabled, invited",
                )

        response = self._http.patch(
            MgmtV1.user_patch_batch_path,
            body=User._compose_patch_batch_body(users, test),
        )
        return response.json()

    def delete(
        self,
        login_id: str,
    ):
        """
        Delete an existing user. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        login_id (str): The login ID of the user to be deleted.

        Raise:
        AuthException: raised if delete operation fails
        """
        self._http.post(
            MgmtV1.user_delete_path,
            body={"loginId": login_id},
        )

    def delete_by_user_id(
        self,
        user_id: str,
    ):
        """
        Delete an existing user by user ID. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        user_id (str): The user ID from the user's JWT.

        Raise:
        AuthException: raised if delete operation fails
        """
        self._http.post(
            MgmtV1.user_delete_path,
            body={"userId": user_id},
        )

    def delete_all_test_users(
        self,
    ):
        """
        Delete all test users in the project. IMPORTANT: This action is irreversible. Use carefully.

        Raise:
        AuthException: raised if delete operation fails
        """
        self._http.delete(
            MgmtV1.user_delete_all_test_users_path,
        )

    def load(
        self,
        login_id: str,
    ) -> dict:
        """
        Load an existing user.

        Args:
        login_id (str): The login ID of the user to be loaded.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the loaded user information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(
            uri=MgmtV1.user_load_path,
            params={"loginId": login_id},
        )
        return response.json()

    def load_by_user_id(
        self,
        user_id: str,
    ) -> dict:
        """
        Load an existing user by user ID.
        The user ID can be found on the user's JWT.

        Args:
        user_id (str): The user ID from the user's JWT.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the loaded user information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(
            uri=MgmtV1.user_load_path,
            params={"userId": user_id},
        )
        return response.json()

    def logout_user(
        self,
        login_id: str,
    ):
        """
        Logout a user from all devices.

        Args:
        login_id (str): The login ID of the user to be logged out.

        Raise:
        AuthException: raised if logout operation fails
        """
        self._http.post(
            MgmtV1.user_logout_path,
            body={"loginId": login_id},
        )

    def logout_user_by_user_id(
        self,
        user_id: str,
    ):
        """
        Logout a user from all devices.

        Args:
        user_id (str): The login ID of the user to be logged out.

        Raise:
        AuthException: raised if logout operation fails
        """
        self._http.post(
            MgmtV1.user_logout_path,
            body={"userId": user_id},
        )

    def load_users(
        self,
        user_ids: List[str],
        include_invalid_users: Optional[bool] = None,
    ) -> dict:
        """
        Load users by their user IDs.

        Args:
        user_ids (List[str]): Optional list of user IDs to filter by
        include_invalid_users (bool): Optional flag to include invalid users in the response

        Return value (dict):
        Return dict in the format
             {"users": []}
        "users" contains a list of all of the found users and their information

        Raise:
        AuthException: raised if search operation fails
        """
        if user_ids is None or len(user_ids) == 0:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "At least one user id needs to be supplied",
            )

        body: dict[str, Union[List[str], bool]] = {
            "userIds": user_ids,
        }

        if include_invalid_users is not None:
            body["includeInvalidUsers"] = include_invalid_users

        response = self._http.post(
            MgmtV1.users_load_path,
            body=body,
        )
        return response.json()

    def search_all(
        self,
        tenant_ids: Optional[List[str]] = None,
        role_names: Optional[List[str]] = None,
        limit: int = 0,
        page: int = 0,
        test_users_only: bool = False,
        with_test_user: bool = False,
        custom_attributes: Optional[dict] = None,
        statuses: Optional[List[str]] = None,
        emails: Optional[List[str]] = None,
        phones: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
        sort: Optional[List[Sort]] = None,
        text: Optional[str] = None,
        login_ids: Optional[List[str]] = None,
        from_created_time: Optional[int] = None,
        to_created_time: Optional[int] = None,
        from_modified_time: Optional[int] = None,
        to_modified_time: Optional[int] = None,
        user_ids: Optional[List[str]] = None,
        tenant_role_ids: Optional[dict] = None,
        tenant_role_names: Optional[dict] = None,
    ) -> dict:
        """
        Search all users.

        Args:
        tenant_ids (List[str]): Optional list of tenant IDs to filter by
        role_names (List[str]): Optional list of role names to filter by
        limit (int): Optional limit of the number of users returned. Leave empty for default.
        page (int): Optional pagination control. Pages start at 0 and must be non-negative.
        test_users_only (bool): Optional filter only test users.
        with_test_user (bool): Optional include test users in search.
        custom_attributes (dict): Optional search for a attribute with a given value
        statuses (List[str]): Optional list of statuses to search for ("enabled", "disabled", "invited")
        emails (List[str]): Optional list of emails to search for
        phones (List[str]): Optional list of phones to search for
        sso_app_ids (List[str]): Optional list of SSO application IDs to filter by
        text (str): Optional string, allows free text search among all user's attributes.
        login_ids (List[str]): Optional list of login ids
        sort (List[Sort]): Optional List[dict], allows to sort by fields.
        from_created_time (int): Optional int, only include users who were created on or after this time (in Unix epoch milliseconds)
        to_created_time (int): Optional int, only include users who were created on or before this time (in Unix epoch milliseconds)
        from_modified_time (int): Optional int, only include users whose last modification/update occurred on or after this time (in Unix epoch milliseconds)
        to_modified_time (int): Optional int, only include users whose last modification/update occurred on or before this time (in Unix epoch milliseconds)
        user_ids (List[str]): Optional list of user IDs to filter by
        tenant_role_ids (dict): Optional mapping of tenant ID to list of role IDs.
            Dict value is in the form of {"tenant_id": {"values":["role_id1", "role_id2"], "and": True}} if you want to match all roles (AND) or any role (OR).
        tenant_role_names (dict): Optional mapping of tenant ID to list of role names.
            Dict value is in the form of {"tenant_id": {"values":["role_name1", "role_name2"], "and": True}} if you want to match all roles (AND) or any role (OR).

        Return value (dict):
        Return dict in the format
             {"users": []}
        "users" contains a list of all of the found users and their information

        Raise:
        AuthException: raised if search operation fails
        """
        tenant_ids = [] if tenant_ids is None else tenant_ids
        role_names = [] if role_names is None else role_names

        if limit < 0:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "limit must be non-negative"
            )

        if page < 0:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "page must be non-negative"
            )
        body = {
            "tenantIds": tenant_ids,
            "roleNames": role_names,
            "limit": limit,
            "page": page,
            "testUsersOnly": test_users_only,
            "withTestUser": with_test_user,
        }
        if statuses is not None:
            body["statuses"] = statuses

        if emails is not None:
            body["emails"] = emails

        if phones is not None:
            body["phones"] = phones

        if custom_attributes is not None:
            body["customAttributes"] = custom_attributes

        if sso_app_ids is not None:
            body["ssoAppIds"] = sso_app_ids

        if login_ids is not None:
            body["loginIds"] = login_ids

        if user_ids is not None:
            body["userIds"] = user_ids

        if text is not None:
            body["text"] = text

        if sort is not None:
            body["sort"] = sort_to_dict(sort)

        if from_created_time is not None:
            body["fromCreatedTime"] = from_created_time
        if to_created_time is not None:
            body["toCreatedTime"] = to_created_time
        if from_modified_time is not None:
            body["fromModifiedTime"] = from_modified_time
        if to_modified_time is not None:
            body["toModifiedTime"] = to_modified_time

        if tenant_role_ids is not None:
            body["tenantRoleIds"] = tenant_role_ids
        if tenant_role_names is not None:
            body["tenantRoleNames"] = tenant_role_names

        response = self._http.post(
            MgmtV1.users_search_path,
            body=body,
        )
        return response.json()

    def search_all_test_users(
        self,
        tenant_ids: Optional[List[str]] = None,
        role_names: Optional[List[str]] = None,
        limit: int = 0,
        page: int = 0,
        custom_attributes: Optional[dict] = None,
        statuses: Optional[List[str]] = None,
        emails: Optional[List[str]] = None,
        phones: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
        sort: Optional[List[Sort]] = None,
        text: Optional[str] = None,
        login_ids: Optional[List[str]] = None,
        from_created_time: Optional[int] = None,
        to_created_time: Optional[int] = None,
        from_modified_time: Optional[int] = None,
        to_modified_time: Optional[int] = None,
        tenant_role_ids: Optional[dict] = None,
        tenant_role_names: Optional[dict] = None,
    ) -> dict:
        """
        Search all test users.

        Args:
        tenant_ids (List[str]): Optional list of tenant IDs to filter by
        role_names (List[str]): Optional list of role names to filter by
        limit (int): Optional limit of the number of users returned. Leave empty for default.
        page (int): Optional pagination control. Pages start at 0 and must be non-negative.
        custom_attributes (dict): Optional search for a attribute with a given value
        statuses (List[str]): Optional list of statuses to search for ("enabled", "disabled", "invited")
        emails (List[str]): Optional list of emails to search for
        phones (List[str]): Optional list of phones to search for
        sso_app_ids (List[str]): Optional list of SSO application IDs to filter by
        text (str): Optional string, allows free text search among all user's attributes.
        login_ids (List[str]): Optional list of login ids
        sort (List[Sort]): Optional List[dict], allows to sort by fields.
        from_created_time (int): Optional int, only include users who were created on or after this time (in Unix epoch milliseconds)
        to_created_time (int): Optional int, only include users who were created on or before this time (in Unix epoch milliseconds)
        from_modified_time (int): Optional int, only include users whose last modification/update occurred on or after this time (in Unix epoch milliseconds)
        to_modified_time (int): Optional int, only include users whose last modification/update occurred on or before this time (in Unix epoch milliseconds)
        tenant_role_ids (dict): Optional mapping of tenant ID to list of role IDs.
            Dict value is in the form of {"tenant_id": {"values":["role_id1", "role_id2"], "and": True}} if you want to match all roles (AND) or any role (OR).
        tenant_role_names (dict): Optional mapping of tenant ID to list of role names.
            Dict value is in the form of {"tenant_id": {"values":["role_name1", "role_name2"], "and": True}} if you want to match all roles (AND) or any role (OR).

        Return value (dict):
        Return dict in the format
             {"users": []}
        "users" contains a list of all of the found users and their information

        Raise:
        AuthException: raised if search operation fails
        """
        tenant_ids = [] if tenant_ids is None else tenant_ids
        role_names = [] if role_names is None else role_names

        if limit < 0:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "limit must be non-negative"
            )

        if page < 0:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "page must be non-negative"
            )
        body = {
            "tenantIds": tenant_ids,
            "roleNames": role_names,
            "limit": limit,
            "page": page,
            "testUsersOnly": True,
            "withTestUser": True,
        }
        if statuses is not None:
            body["statuses"] = statuses

        if emails is not None:
            body["emails"] = emails

        if phones is not None:
            body["phones"] = phones

        if custom_attributes is not None:
            body["customAttributes"] = custom_attributes

        if sso_app_ids is not None:
            body["ssoAppIds"] = sso_app_ids

        if login_ids is not None:
            body["loginIds"] = login_ids

        if text is not None:
            body["text"] = text

        if sort is not None:
            body["sort"] = sort_to_dict(sort)

        if from_created_time is not None:
            body["fromCreatedTime"] = from_created_time
        if to_created_time is not None:
            body["toCreatedTime"] = to_created_time
        if from_modified_time is not None:
            body["fromModifiedTime"] = from_modified_time
        if to_modified_time is not None:
            body["toModifiedTime"] = to_modified_time

        if tenant_role_ids is not None:
            body["tenantRoleIds"] = tenant_role_ids
        if tenant_role_names is not None:
            body["tenantRoleNames"] = tenant_role_names

        response = self._http.post(
            MgmtV1.test_users_search_path,
            body=body,
        )
        return response.json()

    def get_provider_token(
        self,
        login_id: str,
        provider: str,
        withRefreshToken: Optional[bool] = False,
        forceRefresh: Optional[bool] = False,
    ) -> dict:
        """
        Get the provider token for the given login ID.
        Only users that sign-in using social providers will have token.
        Note: The 'Manage tokens from provider' setting must be enabled.

        Args:
        login_id (str): The login ID of the user.
        provider (str): The provider name (google, facebook, etc').
        withRefreshToken (bool): Optional, set to true to get also the refresh token.
        forceRefresh (bool): Optional, set to true to force refresh the token.

        Return value (dict):
        Return dict in the format
             {"provider": "", "providerUserId": "", "accessToken": "", "expiration": "", "scopes": "[]"}
        Containing the provider token of the given user and provider.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.get(
            MgmtV1.user_get_provider_token,
            params={
                "loginId": login_id,
                "provider": provider,
                "withRefreshToken": withRefreshToken,
                "forceRefresh": forceRefresh,
            },
        )
        return response.json()

    def activate(
        self,
        login_id: str,
    ) -> dict:
        """
        Activate an existing user.

        Args:
        login_id (str): The login ID of the user to be activated.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if activate operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_status_path,
            body={"loginId": login_id, "status": "enabled"},
        )
        return response.json()

    def deactivate(
        self,
        login_id: str,
    ) -> dict:
        """
        Deactivate an existing user.

        Args:
        login_id (str): The login ID of the user to be deactivated.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if deactivate operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_status_path,
            body={"loginId": login_id, "status": "disabled"},
        )
        return response.json()

    def update_login_id(
        self,
        login_id: str,
        new_login_id: Optional[str] = None,
    ) -> dict:
        """
        Update login id of user, leave new login empty to remove the ID.
        A user must have at least one login ID. Trying to remove the last one will fail.

        Args:
        login_id (str): The login ID of the user to update.
        new_login_id (str): New login ID to set for the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the update operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_login_id_path,
            body={"loginId": login_id, "newLoginId": new_login_id},
        )
        return response.json()

    def update_email(
        self,
        login_id: str,
        email: Optional[str] = None,
        verified: Optional[bool] = None,
    ) -> dict:
        """
        Update the email address for an existing user.

        Args:
        login_id (str): The login ID of the user to update the email for.
        email (str): The user email address. Leave empty to remove.
        verified (bool): Set to true for the user to be able to login with the email address.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the update operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_email_path,
            body={"loginId": login_id, "email": email, "verified": verified},
        )
        return response.json()

    def update_phone(
        self,
        login_id: str,
        phone: Optional[str] = None,
        verified: Optional[bool] = None,
    ) -> dict:
        """
        Update the phone number for an existing user.

        Args:
        login_id (str): The login ID of the user to update the phone for.
        phone (str): The user phone number. Leave empty to remove.
        verified (bool): Set to true for the user to be able to login with the phone number.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the update operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_phone_path,
            body={"loginId": login_id, "phone": phone, "verified": verified},
        )
        return response.json()

    def update_display_name(
        self,
        login_id: str,
        display_name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
    ) -> dict:
        """
        Update the display name for an existing user.

        Args:
        login_id (str): The login ID of the user to update.
        display_name (str): Optional user display name. Leave empty to remove.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the update operation fails
        """
        bdy = {"loginId": login_id}
        if display_name is not None:
            bdy["displayName"] = display_name
        if given_name is not None:
            bdy["givenName"] = given_name
        if middle_name is not None:
            bdy["middleName"] = middle_name
        if family_name is not None:
            bdy["familyName"] = family_name
        response = self._http.post(
            MgmtV1.user_update_name_path,
            body=bdy,
        )
        return response.json()

    def update_picture(
        self,
        login_id: str,
        picture: Optional[str] = None,
    ) -> dict:
        """
        Update the picture for an existing user.

        Args:
        login_id (str): The login ID of the user to update.
        picture (str): Optional url to user avatar. Leave empty to remove.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the update operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_picture_path,
            body={"loginId": login_id, "picture": picture},
        )
        return response.json()

    def update_custom_attribute(
        self, login_id: str, attribute_key: str, attribute_val: Union[str, int, bool]
    ) -> dict:
        """
        Update a custom attribute of an existing user.

        Args:
        login_id (str): The login ID of the user to update.
        attribute_key (str): The custom attribute that needs to be updated, this attribute needs to exists in Descope console app
        attribute_val: The value to be updated

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the update operation fails
        """
        response = self._http.post(
            MgmtV1.user_update_custom_attribute_path,
            body={
                "loginId": login_id,
                "attributeKey": attribute_key,
                "attributeValue": attribute_val,
            },
        )
        return response.json()

    def set_roles(
        self,
        login_id: str,
        role_names: List[str],
    ) -> dict:
        """
        Set roles to a user without tenant association. Use set_tenant_roles
        for users that are part of a multi-tenant project.

        Args:
        login_id (str): The login ID of the user to update.
        role_names (List[str]): A list of roles to set to a user without tenant association.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_set_role_path,
            body={"loginId": login_id, "roleNames": role_names},
        )
        return response.json()

    def add_roles(
        self,
        login_id: str,
        role_names: List[str],
    ) -> dict:
        """
        Add roles to a user without tenant association. Use add_tenant_roles
        for users that are part of a multi-tenant project.

        Args:
        login_id (str): The login ID of the user to update.
        role_names (List[str]): A list of roles to add to a user without tenant association.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_add_role_path,
            body={"loginId": login_id, "roleNames": role_names},
        )
        return response.json()

    def remove_roles(
        self,
        login_id: str,
        role_names: List[str],
    ) -> dict:
        """
        Remove roles from a user without tenant association. Use remove_tenant_roles
        for users that are part of a multi-tenant project.

        Args:
        login_id (str): The login ID of the user to update.
        role_names (List[str]): A list of roles to remove from a user without tenant association.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_remove_role_path,
            body={"loginId": login_id, "roleNames": role_names},
        )
        return response.json()

    def set_sso_apps(
        self,
        login_id: str,
        sso_app_ids: List[str],
    ) -> dict:
        """
        Set SSO applications association to a user.

        Args:
        login_id (str): The login ID of the user to update.
        sso_app_ids (List[str]): A list of sso applications ids for associate with a user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_set_sso_apps,
            body={"loginId": login_id, "ssoAppIds": sso_app_ids},
        )
        return response.json()

    def add_sso_apps(
        self,
        login_id: str,
        sso_app_ids: List[str],
    ) -> dict:
        """
        Add SSO applications association to a user.

        Args:
        login_id (str): The login ID of the user to update.
        sso_app_ids (List[str]): A list of sso applications ids for associate with a user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_add_sso_apps,
            body={"loginId": login_id, "ssoAppIds": sso_app_ids},
        )
        return response.json()

    def remove_sso_apps(
        self,
        login_id: str,
        sso_app_ids: List[str],
    ) -> dict:
        """
        Remove SSO applications association from a user.

        Args:
        login_id (str): The login ID of the user to update.
        sso_app_ids (List[str]): A list of sso applications ids to remove association from a user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_remove_sso_apps,
            body={"loginId": login_id, "ssoAppIds": sso_app_ids},
        )
        return response.json()

    def add_tenant(
        self,
        login_id: str,
        tenant_id: str,
    ) -> dict:
        """
        Add a tenant association to an existing user.

        Args:
        login_id (str): The login ID of the user to update.
        tenant_id (str): The ID of the tenant to add to the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_add_tenant_path,
            body={"loginId": login_id, "tenantId": tenant_id},
        )
        return response.json()

    def remove_tenant(
        self,
        login_id: str,
        tenant_id: str,
    ) -> dict:
        """
        Remove a tenant association from an existing user.

        Args:
        login_id (str): The login ID of the user to update.
        tenant_id (str): The ID of the tenant to add to the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_remove_tenant_path,
            body={"loginId": login_id, "tenantId": tenant_id},
        )
        return response.json()

    def set_tenant_roles(
        self,
        login_id: str,
        tenant_id: str,
        role_names: List[str],
    ) -> dict:
        """
        Set roles to a user in a specific tenant.

        Args:
        login_id (str): The login ID of the user to update.
        tenant_id (str): The ID of the user's tenant.
        role_names (List[str]): A list of roles to set on the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_set_role_path,
            body={"loginId": login_id, "tenantId": tenant_id, "roleNames": role_names},
        )
        return response.json()

    def add_tenant_roles(
        self,
        login_id: str,
        tenant_id: str,
        role_names: List[str],
    ) -> dict:
        """
        Add roles to a user in a specific tenant.

        Args:
        login_id (str): The login ID of the user to update.
        tenant_id (str): The ID of the user's tenant.
        role_names (List[str]): A list of roles to add to the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_add_role_path,
            body={"loginId": login_id, "tenantId": tenant_id, "roleNames": role_names},
        )
        return response.json()

    def remove_tenant_roles(
        self,
        login_id: str,
        tenant_id: str,
        role_names: List[str],
    ) -> dict:
        """
        Remove roles from a user in a specific tenant.

        Args:
        login_id (str): The login ID of the user to update.
        tenant_id (str): The ID of the user's tenant.
        role_names (List[str]): A list of roles to remove from the user.

        Return value (dict):
        Return dict in the format
             {"user": {}}
        Containing the updated user information.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_remove_role_path,
            body={"loginId": login_id, "tenantId": tenant_id, "roleNames": role_names},
        )
        return response.json()

    def set_temporary_password(
        self,
        login_id: str,
        password: str,
    ) -> None:
        """
            Set the temporary password for the given login ID.
            Note: The password will automatically be set as expired.
            The user will not be able to log-in with this password, and will be required to replace it on next login.
            See also: expire_password

        Args:
        login_id (str): The login ID of the user to set the password to.
        password (str): The new password to set to the user.

        Raise:
        AuthException: raised if the operation fails
        """
        self._http.post(
            MgmtV1.user_set_temporary_password_path,
            body={
                "loginId": login_id,
                "password": password,
                "setActive": False,
            },
        )
        return

    def set_active_password(
        self,
        login_id: str,
        password: str,
    ) -> None:
        """
            Set the password for the given login ID.

        Args:
        login_id (str): The login ID of the user to set the password to.
        password (str): The new password to set to the user.

        Raise:
        AuthException: raised if the operation fails
        """
        self._http.post(
            MgmtV1.user_set_active_password_path,
            body={
                "loginId": login_id,
                "password": password,
                "setActive": True,
            },
        )
        return

    # Deprecated (use set_temporary_password instead)
    def set_password(
        self,
        login_id: str,
        password: str,
        set_active: Optional[bool] = False,
    ) -> None:
        """
            Set the password for the given login ID.
            Note: The password will automatically be set as expired unless the set_active flag will be set to True,
            The user will not be able to log-in with this password, and will be required to replace it on next login.
            See also: expire_password

        Args:
        login_id (str): The login ID of the user to set the password to.
        password (str): The new password to set to the user.
        set_active (bool): Keep the password active so it will not be expired on next log-in

        Raise:
        AuthException: raised if the operation fails
        """
        self._http.post(
            MgmtV1.user_set_password_path,
            body={
                "loginId": login_id,
                "password": password,
                "setActive": set_active,
            },
        )
        return

    def expire_password(
        self,
        login_id: str,
    ) -> None:
        """
            Expires the password for the given login ID.
            Note: user sign-in with an expired password, the user will get an error with code.
            Use the `password.send_reset` or `password.replace` methods to reset/replace the password.

        Args:
        login_id (str): The login ID of the user to expire the password to.

        Raise:
        AuthException: raised if the operation fails
        """
        self._http.post(
            MgmtV1.user_expire_password_path,
            body={"loginId": login_id},
        )
        return

    def remove_all_passkeys(
        self,
        login_id: str,
    ) -> None:
        """
            Removes all registered passkeys (WebAuthn devices) for the user with the given login ID.
            Note: The user might not be able to login anymore if they have no other authentication
            methods or a verified email/phone.

        Args:
        login_id (str): The login ID of the user to remove passkeys for.

        Raise:
        AuthException: raised if the operation fails
        """
        self._http.post(
            MgmtV1.user_remove_all_passkeys_path,
            body={"loginId": login_id},
        )
        return

    def remove_totp_seed(
        self,
        login_id: str,
    ) -> None:
        """
            Removes TOTP seed for the user with the given login ID.
            Note: The user might not be able to login anymore if they have no other authentication
            methods or a verified email/phone.

        Args:
        login_id (str): The login ID of the user to remove totp seed for.

        Raise:
        AuthException: raised if the operation fails
        """
        self._http.post(
            MgmtV1.user_remove_totp_seed_path,
            body={"loginId": login_id},
        )
        return

    def generate_otp_for_test_user(
        self,
        method: DeliveryMethod,
        login_id: str,
        login_options: Optional[LoginOptions] = None,
    ) -> dict:
        """
        Generate OTP for the given login ID of a test user.
        This is useful when running tests and don't want to use 3rd party messaging services.

        Args:
        method (DeliveryMethod): The method to use for "delivering" the OTP verification code to the user, for example
            EMAIL, SMS, VOICE, WHATSAPP or EMBEDDED
        login_id (str): The login ID of the test user being validated.
        login_options (LoginOptions): optional, can be provided to set custom claims to the generated jwt.

        Return value (dict):
        Return dict in the format
             {"code": "", "loginId": ""}
        Containing the code for the login (exactly as it sent via Email or Phone messaging).

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_generate_otp_for_test_path,
            body={
                "loginId": login_id,
                "deliveryMethod": get_method_string(method),
                "loginOptions": login_options.__dict__ if login_options else {},
            },
        )
        return response.json()

    def generate_magic_link_for_test_user(
        self,
        method: DeliveryMethod,
        login_id: str,
        uri: str,
        login_options: Optional[LoginOptions] = None,
    ) -> dict:
        """
        Generate Magic Link for the given login ID of a test user.
        This is useful when running tests and don't want to use 3rd party messaging services.

        Args:
        method (DeliveryMethod): The method to use for "delivering" the verification magic link to the user, for example
            EMAIL, SMS, VOICE, WHATSAPP or EMBEDDED
        login_id (str): The login ID of the test user being validated.
        uri (str): Optional redirect uri which will be used instead of any global configuration.
        login_options (LoginOptions): optional, can be provided to set custom claims to the generated jwt.

        Return value (dict):
        Return dict in the format
             {"link": "", "loginId": ""}
        Containing the magic link for the login (exactly as it sent via Email or Phone messaging).

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_generate_magic_link_for_test_path,
            body={
                "loginId": login_id,
                "deliveryMethod": get_method_string(method),
                "URI": uri,
                "loginOptions": login_options.__dict__ if login_options else {},
            },
        )
        return response.json()

    def generate_enchanted_link_for_test_user(
        self,
        login_id: str,
        uri: str,
        login_options: Optional[LoginOptions] = None,
    ) -> dict:
        """
        Generate Enchanted Link for the given login ID of a test user.
        This is useful when running tests and don't want to use 3rd party messaging services.

        Args:
        login_id (str): The login ID of the test user being validated.
        uri (str): Optional redirect uri which will be used instead of any global configuration.
        login_options (LoginOptions): optional, can be provided to set custom claims to the generated jwt.

        Return value (dict):
        Return dict in the format
             {"link": "", "loginId": "", "pendingRef": ""}
        Containing the enchanted link for the login (exactly as it sent via Email or Phone messaging) and pendingRef.

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_generate_enchanted_link_for_test_path,
            body={
                "loginId": login_id,
                "URI": uri,
                "loginOptions": login_options.__dict__ if login_options else {},
            },
        )
        return response.json()

    def generate_embedded_link(
        self, login_id: str, custom_claims: Optional[dict] = None, timeout: int = 0
    ) -> str:
        """
        Generate Embedded Link for the given user login ID.
        The return value is a token that can be verified via magic link, or using flows

        Args:
        login_id (str): The login ID of the user to authenticate with.
        custom_claims (dict): Additional claims to place on the jwt after verification

        Return value (str):
        Return the token to be used in verification process

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_generate_embedded_link_path,
            body={
                "loginId": login_id,
                "customClaims": custom_claims,
                "timeout": timeout,
            },
        )
        return response.json()["token"]

    def generate_sign_up_embedded_link(
        self,
        login_id: str,
        user: Optional[CreateUserObj] = None,
        email_verified: bool = False,
        phone_verified: bool = False,
        login_options: Optional[LoginOptions] = None,
        timeout: int = 0,
    ) -> str:
        """
        Generate sign up Embedded Link for the given user login ID.
        The return value is a token that can be verified via magic link, or using flows

        Args:
        login_id (str): The login ID of the user to authenticate with.
        user (CreateUserObj): Optional user object to create the user with
        email_verified (bool): Optional, set to true if the email is verified
        phone_verified (bool): Optional, set to true if the phone is verified
        login_options (LoginOptions): Optional login options to customize the link
        timeout (int): Optional, the timeout in seconds for the link to be valid

        Return value (str):
        Return the token to be used in verification process

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_generate_sign_up_embedded_link_path,
            body={
                "loginId": login_id,
                "user": user.__dict__ if user else {},
                "loginOptions": login_options.__dict__ if login_options else {},
                "emailVerified": email_verified,
                "phoneVerified": phone_verified,
                "timeout": timeout,
            },
        )
        return response.json()["token"]

    def history(self, user_ids: List[str]) -> List[dict]:
        """
        Retrieve users' authentication history, by the given user's ids.

        Args:
        login_ids (List[str]): List of Users' IDs.

        Return value (List[dict]):
        Return List in the format
             [
                {
                    "userId": "User's ID",
                    "loginTime": "User'sLogin time",
                    "city": "User's city",
                    "country": "User's country",
                    "ip": User's IP
                }
            ]

        Raise:
        AuthException: raised if the operation fails
        """
        response = self._http.post(
            MgmtV1.user_history_path,
            body=user_ids,
        )
        return response.json()

    @staticmethod
    def _compose_create_body(
        login_id: str,
        email: Optional[str],
        phone: Optional[str],
        display_name: Optional[str],
        given_name: Optional[str],
        middle_name: Optional[str],
        family_name: Optional[str],
        role_names: List[str],
        user_tenants: List[AssociatedTenant],
        invite: bool,
        test: bool,
        picture: Optional[str],
        custom_attributes: Optional[dict],
        verified_email: Optional[bool],
        verified_phone: Optional[bool],
        invite_url: Optional[str],
        send_mail: Optional[bool],
        send_sms: Optional[bool],
        additional_login_ids: Optional[List[str]],
        sso_app_ids: Optional[List[str]] = None,
        template_id: str = "",
    ) -> dict:
        body = User._compose_update_body(
            login_id=login_id,
            email=email,
            phone=phone,
            display_name=display_name,
            given_name=given_name,
            middle_name=middle_name,
            family_name=family_name,
            role_names=role_names,
            user_tenants=user_tenants,
            test=test,
            picture=picture,
            custom_attributes=custom_attributes,
            additional_login_ids=additional_login_ids,
            sso_app_ids=sso_app_ids,
        )
        body["invite"] = invite
        if verified_email is not None:
            body["verifiedEmail"] = verified_email
        if verified_phone is not None:
            body["verifiedPhone"] = verified_phone
        if invite_url is not None:
            body["inviteUrl"] = invite_url
        if send_mail is not None:
            body["sendMail"] = send_mail
        if send_sms is not None:
            body["sendSMS"] = send_sms
        if template_id != "":
            body["templateId"] = template_id
        return body

    @staticmethod
    def _compose_create_batch_body(
        users: List[UserObj],
        invite_url: Optional[str],
        send_mail: Optional[bool],
        send_sms: Optional[bool],
    ) -> dict:
        usersBody = []
        for user in users:
            role_names = [] if user.role_names is None else user.role_names
            user_tenants = [] if user.user_tenants is None else user.user_tenants
            sso_app_ids = [] if user.sso_app_ids is None else user.sso_app_ids
            password = None if user.password is None else user.password.cleartext
            hashed_password = None
            if (user.password is not None) and (user.password.hashed is not None):
                hashed_password = user.password.hashed.to_dict()
            uBody = User._compose_update_body(
                login_id=user.login_id,
                email=user.email,
                phone=user.phone,
                display_name=user.display_name,
                given_name=user.given_name,
                middle_name=user.middle_name,
                family_name=user.family_name,
                role_names=role_names,
                user_tenants=user_tenants,
                picture=user.picture,
                custom_attributes=user.custom_attributes,
                additional_login_ids=user.additional_login_ids,
                verified_email=user.verified_email,
                verified_phone=user.verified_phone,
                test=False,
                sso_app_ids=sso_app_ids,
                password=password,
                hashed_password=hashed_password,
                seed=user.seed,
            )
            if user.status is not None:
                uBody["status"] = user.status
            usersBody.append(uBody)

        body = {"users": usersBody, "invite": True}
        if invite_url is not None:
            body["inviteUrl"] = invite_url
        if send_mail is not None:
            body["sendMail"] = send_mail
        if send_sms is not None:
            body["sendSMS"] = send_sms
        return body

    @staticmethod
    def _compose_update_body(
        login_id: str,
        email: Optional[str],
        phone: Optional[str],
        display_name: Optional[str],
        given_name: Optional[str],
        middle_name: Optional[str],
        family_name: Optional[str],
        role_names: List[str],
        user_tenants: List[AssociatedTenant],
        test: bool,
        picture: Optional[str],
        custom_attributes: Optional[dict],
        verified_email: Optional[bool] = None,
        verified_phone: Optional[bool] = None,
        additional_login_ids: Optional[List[str]] = None,
        sso_app_ids: Optional[List[str]] = None,
        password: Optional[str] = None,
        hashed_password: Optional[dict] = None,
        seed: Optional[str] = None,
    ) -> dict:
        res = {
            "loginId": login_id,
            "email": email,
            "phone": phone,
            "displayName": display_name,
            "roleNames": role_names,
            "userTenants": associated_tenants_to_dict(user_tenants),
            "test": test,
            "picture": picture,
            "customAttributes": custom_attributes,
            "additionalLoginIds": additional_login_ids,
            "ssoAppIDs": sso_app_ids,
        }
        if verified_email is not None:
            res["verifiedEmail"] = verified_email
        if given_name is not None:
            res["givenName"] = given_name
        if middle_name is not None:
            res["middleName"] = middle_name
        if family_name is not None:
            res["familyName"] = family_name
        if verified_phone is not None:
            res["verifiedPhone"] = verified_phone
        if password is not None:
            res["password"] = password
        if hashed_password is not None:
            res["hashedPassword"] = hashed_password
        if seed is not None:
            res["seed"] = seed
        return res

    @staticmethod
    def _compose_patch_body(
        login_id: str,
        email: Optional[str],
        phone: Optional[str],
        display_name: Optional[str],
        given_name: Optional[str],
        middle_name: Optional[str],
        family_name: Optional[str],
        role_names: Optional[List[str]],
        user_tenants: Optional[List[AssociatedTenant]],
        picture: Optional[str],
        custom_attributes: Optional[dict],
        verified_email: Optional[bool],
        verified_phone: Optional[bool],
        sso_app_ids: Optional[List[str]],
        status: Optional[str],
        test: bool = False,
    ) -> dict:
        res: dict[str, Any] = {
            "loginId": login_id,
        }
        if email is not None:
            res["email"] = email
        if phone is not None:
            res["phone"] = phone
        if display_name is not None:
            res["displayName"] = display_name
        if given_name is not None:
            res["givenName"] = given_name
        if middle_name is not None:
            res["middleName"] = middle_name
        if family_name is not None:
            res["familyName"] = family_name
        if role_names is not None:
            res["roleNames"] = role_names
        if user_tenants is not None:
            res["userTenants"] = associated_tenants_to_dict(user_tenants)
        if picture is not None:
            res["picture"] = picture
        if custom_attributes is not None:
            res["customAttributes"] = custom_attributes
        if verified_email is not None:
            res["verifiedEmail"] = verified_email
        if verified_phone is not None:
            res["verifiedPhone"] = verified_phone
        if sso_app_ids is not None:
            res["ssoAppIds"] = sso_app_ids
        if status is not None:
            res["status"] = status
        if test:
            res["test"] = test
        return res

    @staticmethod
    def _compose_patch_batch_body(
        users: List[UserObj],
        test: bool = False,
    ) -> dict:
        users_body = []
        for user in users:
            user_body = User._compose_patch_body(
                login_id=user.login_id,
                email=user.email,
                phone=user.phone,
                display_name=user.display_name,
                given_name=user.given_name,
                middle_name=user.middle_name,
                family_name=user.family_name,
                role_names=user.role_names,
                user_tenants=user.user_tenants,
                picture=user.picture,
                custom_attributes=user.custom_attributes,
                verified_email=user.verified_email,
                verified_phone=user.verified_phone,
                sso_app_ids=user.sso_app_ids,
                status=user.status,
                test=test,
            )
            users_body.append(user_body)

        return {"users": users_body}



# This is not part of the public API but a code helper
from descope.auth import Auth


class AuthBase:
    """Base class for classes having auth"""

    def __init__(self, auth: Auth):
        self._auth = auth
        self._http = auth.http_client



from __future__ import annotations

import copy
import json
import os
import re
from http import HTTPStatus
from threading import Lock
from typing import Iterable, Optional

import jwt
from email_validator import EmailNotValidError, validate_email
from jwt import ExpiredSignatureError, ImmatureSignatureError

from descope.common import (
    DEFAULT_BASE_URL,
    DEFAULT_DOMAIN,
    DEFAULT_URL_PREFIX,
    PHONE_REGEX,
    REFRESH_SESSION_COOKIE_NAME,
    SESSION_TOKEN_NAME,
    AccessKeyLoginOptions,
    DeliveryMethod,
    EndpointsV1,
    EndpointsV2,
)
from descope.exceptions import (
    API_RATE_LIMIT_RETRY_AFTER_HEADER,
    ERROR_TYPE_API_RATE_LIMIT,
    ERROR_TYPE_INVALID_ARGUMENT,
    ERROR_TYPE_INVALID_PUBLIC_KEY,
    ERROR_TYPE_INVALID_TOKEN,
    ERROR_TYPE_SERVER_ERROR,
    AuthException,
    RateLimitException,
)
from descope.http_client import HTTPClient
from descope.jwt_common import adjust_properties as jwt_adjust_properties
from descope.jwt_common import generate_auth_info as jwt_generate_auth_info
from descope.jwt_common import generate_jwt_response as jwt_generate_jwt_response


class Auth:
    ALGORITHM_KEY = "alg"

    def __init__(
        self,
        project_id: Optional[str] = None,
        public_key: Optional[dict | str] = None,
        jwt_validation_leeway: int = 5,
        *,
        http_client: HTTPClient,
    ):
        self.lock_public_keys = Lock()

        if not project_id:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                (
                    "Unable to init Auth object because project_id cannot be empty. "
                    "Set environment variable DESCOPE_PROJECT_ID or pass your Project ID to the init function."
                ),
            )
        self.project_id = project_id
        self.jwt_validation_leeway = jwt_validation_leeway

        self._http = http_client

        public_key = public_key or os.getenv("DESCOPE_PUBLIC_KEY")
        with self.lock_public_keys:
            if not public_key:
                self.public_keys = {}
            else:
                kid, pub_key, alg = self._validate_and_load_public_key(public_key)
                self.public_keys = {kid: (pub_key, alg)}

    def _raise_rate_limit_exception(self, response):
        try:
            resp = response.json()
            raise RateLimitException(
                resp.get("errorCode", HTTPStatus.TOO_MANY_REQUESTS),
                ERROR_TYPE_API_RATE_LIMIT,
                resp.get("errorDescription", ""),
                resp.get("errorMessage", ""),
                rate_limit_parameters={
                    API_RATE_LIMIT_RETRY_AFTER_HEADER: self._parse_retry_after(
                        response.headers
                    )
                },
            )
        except RateLimitException:
            raise
        except Exception:
            raise RateLimitException(
                status_code=HTTPStatus.TOO_MANY_REQUESTS,
                error_type=ERROR_TYPE_API_RATE_LIMIT,
                error_message=ERROR_TYPE_API_RATE_LIMIT,
                error_description=ERROR_TYPE_API_RATE_LIMIT,
            )

    def _parse_retry_after(self, headers):
        try:
            return int(headers.get(API_RATE_LIMIT_RETRY_AFTER_HEADER, 0))
        except (ValueError, TypeError):
            return 0

    @property
    def http_client(self) -> HTTPClient:
        return self._http

    def exchange_token(
        self, uri, code: str, audience: Optional[Iterable[str] | str] = None
    ) -> dict:
        if not code:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "exchange code is empty",
            )

        body = Auth._compose_exchange_body(code)
        response = self._http.post(uri, body=body)
        resp = response.json()
        jwt_response = self.generate_jwt_response(
            resp, response.cookies.get(REFRESH_SESSION_COOKIE_NAME), audience
        )
        return jwt_response

    @staticmethod
    def base_url_for_project_id(project_id):
        if len(project_id) >= 32:
            region = project_id[1:5]
            return ".".join([DEFAULT_URL_PREFIX, region, DEFAULT_DOMAIN])
        return DEFAULT_BASE_URL

    @staticmethod
    def adjust_and_verify_delivery_method(
        method: DeliveryMethod, login_id: str, user: dict
    ) -> bool:
        if not login_id:
            return False

        if not isinstance(user, dict):
            return False

        if method == DeliveryMethod.EMAIL:
            if not user.get("email", None):
                user["email"] = login_id
            try:
                validate_email(user["email"], check_deliverability=False)
                return True
            except EmailNotValidError:
                return False
        elif method == DeliveryMethod.SMS:
            if not user.get("phone", None):
                user["phone"] = login_id
            if not re.match(PHONE_REGEX, user["phone"]):
                return False
        elif method == DeliveryMethod.VOICE:
            if not user.get("phone", None):
                user["phone"] = login_id
            if not re.match(PHONE_REGEX, user["phone"]):
                return False
        elif method == DeliveryMethod.WHATSAPP:
            if not user.get("phone", None):
                user["phone"] = login_id
            if not re.match(PHONE_REGEX, user["phone"]):
                return False
        else:
            return False

        return True

    @staticmethod
    def compose_url(base: str, method: DeliveryMethod) -> str:
        suffix = {
            DeliveryMethod.EMAIL: "email",
            DeliveryMethod.SMS: "sms",
            DeliveryMethod.VOICE: "voice",
            DeliveryMethod.WHATSAPP: "whatsapp",
        }.get(method)

        if not suffix:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, f"Unknown delivery method: {method}"
            )

        return f"{base}/{suffix}"

    @staticmethod
    def get_login_id_by_method(method: DeliveryMethod, user: dict) -> tuple[str, str]:
        login_id = {
            DeliveryMethod.EMAIL: ("email", user.get("email", "")),
            DeliveryMethod.SMS: ("phone", user.get("phone", "")),
            DeliveryMethod.VOICE: ("voice", user.get("phone", "")),
            DeliveryMethod.WHATSAPP: ("whatsapp", user.get("phone", "")),
        }.get(method)

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, f"Unknown delivery method: {method}"
            )

        return login_id

    @staticmethod
    def validate_email(email: str):
        if email == "":
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "email address argument cannot be empty",
            )

        try:
            validate_email(email, check_deliverability=False)
        except EmailNotValidError as ex:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, f"Invalid email address: {ex}"
            )

    @staticmethod
    def validate_phone(method: DeliveryMethod, phone: str):
        if phone == "":
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "Phone number argument cannot be empty",
            )

        if not re.match(PHONE_REGEX, phone):
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Invalid phone number"
            )

        if (
            method != DeliveryMethod.SMS
            and method != DeliveryMethod.VOICE
            and method != DeliveryMethod.WHATSAPP
        ):
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Invalid delivery method"
            )

    def exchange_access_key(
        self,
        access_key: str,
        audience: str | Iterable[str] | None = None,
        login_options: AccessKeyLoginOptions | None = None,
    ) -> dict:
        uri = EndpointsV1.exchange_auth_access_key_path
        body = {
            "loginOptions": login_options.__dict__ if login_options else {},
        }
        server_response = self._http.post(uri, body=body, pswd=access_key)
        json_body = server_response.json()
        return self._generate_auth_info(
            response_body=json_body,
            refresh_token=None,
            user_jwt=False,
            audience=audience,
        )

    @staticmethod
    def _compose_exchange_body(code: str) -> dict:
        return {"code": code}

    @staticmethod
    def _validate_and_load_public_key(public_key) -> tuple[str, jwt.PyJWK, str]:
        if not isinstance(public_key, (str, dict)):
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_PUBLIC_KEY,
                "Unable to load public key. Invalid public key error: (unknown type)",
            )

        if isinstance(public_key, str):
            try:
                public_key = json.loads(public_key)
            except (ValueError, TypeError) as e:
                raise AuthException(
                    500,
                    ERROR_TYPE_INVALID_PUBLIC_KEY,
                    f"Unable to load public key. error: {e}",
                )

        alg = public_key.get(Auth.ALGORITHM_KEY, None)
        if alg is None:
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_PUBLIC_KEY,
                "Unable to load public key. Missing property: alg",
            )

        kid = public_key.get("kid", None)
        if kid is None:
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_PUBLIC_KEY,
                "Unable to load public key. Missing property: kid",
            )
        try:
            # Load and validate public key
            return (kid, jwt.PyJWK(public_key), alg)
        except (jwt.PyJWKError, jwt.InvalidKeyError) as e:
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_PUBLIC_KEY,
                f"Unable to load public key {e}",
            )

    def _raise_from_response(self, response):
        """Raise appropriate exception from response, does nothing if response.ok is True."""
        if response.ok:
            return

        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            self._raise_rate_limit_exception(response)  # Raise RateLimitException

        raise AuthException(
            response.status_code,
            ERROR_TYPE_SERVER_ERROR,
            response.text,
        )

    def _fetch_public_keys(self) -> None:
        # This function called under mutex protection so no need to acquire it once again
        response = self._http.get(
            f"{EndpointsV2.public_key_path}/{self.project_id}",
        )

        jwks_data = response.text
        try:
            jwkeys_wrapper = json.loads(jwks_data)
            jwkeys = jwkeys_wrapper["keys"]
        except (ValueError, TypeError, KeyError) as e:
            raise AuthException(
                500, ERROR_TYPE_INVALID_PUBLIC_KEY, f"Unable to load jwks. Error: {e}"
            )

        # Load all public keys for this project
        self.public_keys = {}
        for key in jwkeys:
            try:
                loaded_kid, pub_key, alg = Auth._validate_and_load_public_key(key)
                self.public_keys[loaded_kid] = (pub_key, alg)
            except AuthException:
                # just continue to the next key
                pass

    def adjust_properties(self, jwt_response: dict, user_jwt: bool):
        # Delegate to shared JWT utilities for normalization
        return jwt_adjust_properties(jwt_response, user_jwt)

    def _generate_auth_info(
        self,
        response_body: dict,
        refresh_token: str | None,
        user_jwt: bool,
        audience: str | None | Iterable[str] = None,
    ) -> dict:
        # Use shared generator with class validator to preserve signature checks
        return jwt_generate_auth_info(
            response_body,
            refresh_token,
            user_jwt,
            audience,
            token_validator=self._validate_token,
        )

    def generate_jwt_response(
        self,
        response_body: dict,
        refresh_cookie: str | None,
        audience: str | None | Iterable[str] = None,
    ) -> dict:
        # Delegate to shared implementation (keeps same output shape)
        return jwt_generate_jwt_response(
            response_body,
            refresh_cookie,
            audience,
            token_validator=self._validate_token,
        )

    # public method to validate a token from the management class
    def validate_token(
        self, token: str, audience: str | None | Iterable[str] = None
    ) -> dict:
        return self._validate_token(token, audience)

    # Validate a token and load the public key if needed
    def _validate_token(
        self, token: str, audience: str | None | Iterable[str] = None
    ) -> dict:
        if not token:
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_TOKEN,
                "Token validation received empty token",
            )
        try:
            unverified_header = jwt.get_unverified_header(token)
        except jwt.exceptions.PyJWTError as e:
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_TOKEN,
                f"Unable to parse token header. Error: {e}",
            ) from e

        alg_header = unverified_header.get(Auth.ALGORITHM_KEY, None)
        if alg_header is None or alg_header == "none":
            raise AuthException(
                500, ERROR_TYPE_INVALID_TOKEN, "Token header is missing property: alg"
            )

        kid = unverified_header.get("kid", None)
        if kid is None:
            raise AuthException(
                500, ERROR_TYPE_INVALID_TOKEN, "Token header is missing property: kid"
            )

        with self.lock_public_keys:
            if self.public_keys == {} or self.public_keys.get(kid, None) is None:
                self._fetch_public_keys()

            found_key = self.public_keys.get(kid, None)
            if found_key is None:
                raise AuthException(
                    500,
                    ERROR_TYPE_INVALID_PUBLIC_KEY,
                    "Unable to validate public key. Public key not found.",
                )
            # save reference to the founded key
            # (as another thread can change the self.public_keys dict)
            copy_key = found_key

        alg_from_key = copy_key[1]
        if alg_header != alg_from_key:
            raise AuthException(
                500,
                ERROR_TYPE_INVALID_PUBLIC_KEY,
                "Algorithm signature in JWT header does not match the algorithm signature in the public key",
            )

        # Check if we need to auto-detect audience from token
        validation_audience = audience
        if audience is None:
            try:
                unverified_claims = jwt.decode(
                    jwt=token,
                    key=copy_key[0].key,
                    algorithms=[alg_header],
                    options={"verify_aud": False},  # Skip audience verification for now
                    leeway=self.jwt_validation_leeway,
                )
                token_audience = unverified_claims.get("aud")

                # If token has audience claim and it matches our project ID, use it
                if token_audience and self.project_id:
                    if isinstance(token_audience, list):
                        if self.project_id in token_audience:
                            validation_audience = self.project_id
                    else:
                        if token_audience == self.project_id:
                            validation_audience = self.project_id
            except Exception:
                # If we can't decode the token to check audience, proceed with original audience (None)
                pass

        try:
            claims = jwt.decode(
                jwt=token,
                key=copy_key[0].key,
                algorithms=[alg_header],
                audience=validation_audience,
                leeway=self.jwt_validation_leeway,
            )
        except ImmatureSignatureError:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_TOKEN,
                "Received Invalid token (nbf in future) during jwt validation. Error can be due to time glitch (between machines), try to set the jwt_validation_leeway parameter (in DescopeClient) to higher value than 5sec which is the default",
            )
        except ExpiredSignatureError:
            raise AuthException(
                401,
                ERROR_TYPE_INVALID_TOKEN,
                "Received expired token (exp in past) during jwt validation. (sometimes can be due to time glitch (between machines), try to set the jwt_validation_leeway parameter (in DescopeClient) to higher value than 5sec which is the default)",
            )

        claims["jwt"] = token
        return claims

    def validate_session(
        self, session_token: str, audience: str | None | Iterable[str] = None
    ) -> dict:
        if not session_token:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_TOKEN,
                "Session token is required for validation",
            )

        res = self._validate_token(session_token, audience)
        res[SESSION_TOKEN_NAME] = copy.deepcopy(
            res
        )  # Duplicate for saving backward compatibility but keep the same structure as the refresh operation response
        return self.adjust_properties(res, True)

    def refresh_session(
        self, refresh_token: str, audience: str | None | Iterable[str] = None
    ) -> dict:
        if not refresh_token:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_TOKEN,
                "Refresh token is required to refresh a session",
            )

        self._validate_token(refresh_token, audience)

        uri = EndpointsV1.refresh_token_path
        response = self._http.post(uri, body={}, pswd=refresh_token)

        resp = response.json()
        refresh_token = (
            response.cookies.get(REFRESH_SESSION_COOKIE_NAME, None) or refresh_token
        )
        return self.generate_jwt_response(resp, refresh_token, audience)

    def validate_and_refresh_session(
        self,
        session_token: str,
        refresh_token: str,
        audience: str | None | Iterable[str] = None,
    ) -> dict:
        if not session_token:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_TOKEN,
                "Session token is missing",
            )

        try:
            return self.validate_session(session_token, audience)
        except AuthException:
            # Session is invalid - try to refresh it
            if not refresh_token:
                raise AuthException(
                    400,
                    ERROR_TYPE_INVALID_TOKEN,
                    "Refresh token is missing",
                )
            return self.refresh_session(refresh_token, audience)

    def select_tenant(
        self,
        tenant_id: str,
        refresh_token: str,
        audience: str | None | Iterable[str] = None,
    ) -> dict:
        if not refresh_token:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_TOKEN,
                "Refresh token is required to refresh a session",
            )

        uri = EndpointsV1.select_tenant_path
        response = self._http.post(uri, body={"tenant": tenant_id}, pswd=refresh_token)

        resp = response.json()
        jwt_response = self.generate_jwt_response(
            resp, response.cookies.get(REFRESH_SESSION_COOKIE_NAME, None), audience
        )
        return jwt_response

    @staticmethod
    def extract_masked_address(response: dict, method: DeliveryMethod) -> str:
        if (
            method == DeliveryMethod.SMS
            or method == DeliveryMethod.VOICE
            or method == DeliveryMethod.WHATSAPP
        ):
            return response["maskedPhone"]
        elif method == DeliveryMethod.EMAIL:
            return response["maskedEmail"]
        return ""



from enum import Enum
from typing import Any, Optional

from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException

DEFAULT_URL_PREFIX = "https://api"
DEFAULT_DOMAIN = "descope.com"
DEFAULT_BASE_URL = DEFAULT_URL_PREFIX + "." + DEFAULT_DOMAIN  # pragma: no cover
DEFAULT_TIMEOUT_SECONDS = 60

PHONE_REGEX = r"""^(?:(?:\(?(?:00|\+)([1-4]\d\d|[1-9]\d?)\)?)?[\-\.\ \\/]?)?((?:\(?\d{1,}\)?[\-\.\ \\/]?){0,})(?:[\-\.\ \\/]?(?:#|ext\.?|extension|x)[\-\.\ \\/]?(\d+))?$"""

SESSION_COOKIE_NAME = "DS"
REFRESH_SESSION_COOKIE_NAME = "DSR"

SESSION_TOKEN_NAME = "sessionToken"
REFRESH_SESSION_TOKEN_NAME = "refreshSessionToken"
COOKIE_DATA_NAME = "cookieData"

REDIRECT_LOCATION_COOKIE_NAME = "Location"


class EndpointsV1:
    refresh_token_path = "/v1/auth/refresh"
    select_tenant_path = "/v1/auth/tenant/select"
    logout_path = "/v1/auth/logout"
    logout_all_path = "/v1/auth/logoutall"
    me_path = "/v1/auth/me"
    my_tenants_path = "/v1/auth/me/tenants"
    history_path = "/v1/auth/me/history"

    # accesskey
    exchange_auth_access_key_path = "/v1/auth/accesskey/exchange"

    # otp
    sign_up_auth_otp_path = "/v1/auth/otp/signup"
    sign_in_auth_otp_path = "/v1/auth/otp/signin"
    sign_up_or_in_auth_otp_path = "/v1/auth/otp/signup-in"
    verify_code_auth_path = "/v1/auth/otp/verify"
    update_user_email_otp_path = "/v1/auth/otp/update/email"
    update_user_phone_otp_path = "/v1/auth/otp/update/phone"

    # magiclink
    sign_up_auth_magiclink_path = "/v1/auth/magiclink/signup"
    sign_in_auth_magiclink_path = "/v1/auth/magiclink/signin"
    sign_up_or_in_auth_magiclink_path = "/v1/auth/magiclink/signup-in"
    verify_magiclink_auth_path = "/v1/auth/magiclink/verify"
    get_session_magiclink_auth_path = "/v1/auth/magiclink/pending-session"
    update_user_email_magiclink_path = "/v1/auth/magiclink/update/email"
    update_user_phone_magiclink_path = "/v1/auth/magiclink/update/phone"

    # enchantedlink
    sign_up_auth_enchantedlink_path = "/v1/auth/enchantedlink/signup"
    sign_in_auth_enchantedlink_path = "/v1/auth/enchantedlink/signin"
    sign_up_or_in_auth_enchantedlink_path = "/v1/auth/enchantedlink/signup-in"
    verify_enchantedlink_auth_path = "/v1/auth/enchantedlink/verify"
    get_session_enchantedlink_auth_path = "/v1/auth/enchantedlink/pending-session"
    update_user_email_enchantedlink_path = "/v1/auth/enchantedlink/update/email"

    # oauth
    oauth_start_path = "/v1/auth/oauth/authorize"
    oauth_exchange_token_path = "/v1/auth/oauth/exchange"

    # #### DEPRECATED
    # saml
    auth_saml_start_path = "/v1/auth/saml/authorize"
    saml_exchange_token_path = "/v1/auth/saml/exchange"
    #################

    # sso (saml/oidc)
    auth_sso_start_path = "/v1/auth/sso/authorize"
    sso_exchange_token_path = "/v1/auth/sso/exchange"

    # totp
    sign_up_auth_totp_path = "/v1/auth/totp/signup"
    verify_totp_path = "/v1/auth/totp/verify"
    update_totp_path = "/v1/auth/totp/update"

    # webauthn
    sign_up_auth_webauthn_start_path = "/v1/auth/webauthn/signup/start"
    sign_up_auth_webauthn_finish_path = "/v1/auth/webauthn/signup/finish"
    sign_in_auth_webauthn_start_path = "/v1/auth/webauthn/signin/start"
    sign_in_auth_webauthn_finish_path = "/v1/auth/webauthn/signin/finish"
    sign_up_or_in_auth_webauthn_start_path = "/v1/auth/webauthn/signup-in/start"
    update_auth_webauthn_start_path = "/v1/auth/webauthn/update/start"
    update_auth_webauthn_finish_path = "/v1/auth/webauthn/update/finish"

    # password
    sign_up_password_path = "/v1/auth/password/signup"
    sign_in_password_path = "/v1/auth/password/signin"
    send_reset_password_path = "/v1/auth/password/reset"
    update_password_path = "/v1/auth/password/update"
    replace_password_path = "/v1/auth/password/replace"
    password_policy_path = "/v1/auth/password/policy"


class EndpointsV2:
    public_key_path = "/v2/keys"


class DeliveryMethod(Enum):
    WHATSAPP = 1
    SMS = 2
    EMAIL = 3
    EMBEDDED = 4
    VOICE = 5


class LoginOptions:
    def __init__(
        self,
        stepup: bool = False,
        mfa: bool = False,
        revoke_other_sessions: Optional[bool] = None,
        custom_claims: Optional[dict] = None,
        template_options: Optional[
            dict
        ] = None,  # for providing messaging template options (templates that are being sent via email / text message)
        template_id: Optional[
            str
        ] = None,  # for overriding the default template (templates that are being sent via email / text message)
    ):
        self.stepup = stepup
        self.customClaims = custom_claims
        self.mfa = mfa
        if revoke_other_sessions is not None:
            self.revokeOtherSessions = revoke_other_sessions
        if template_options is not None:
            self.templateOptions = template_options
        if template_id is not None:
            self.templateId = template_id


class AccessKeyLoginOptions:
    def __init__(
        self,
        custom_claims: Optional[dict] = None,
    ):
        self.customClaims = custom_claims


def validate_refresh_token_provided(
    login_options: Optional[LoginOptions] = None, refresh_token: Optional[str] = None
):
    refresh_required = login_options is not None and (
        login_options.mfa or login_options.stepup
    )
    refresh_missing = refresh_token is None or refresh_token == ""
    if refresh_required and refresh_missing:
        raise AuthException(
            400,
            ERROR_TYPE_INVALID_ARGUMENT,
            "Missing refresh token for stepup/mfa",
        )


class SignUpOptions:
    def __init__(
        self,
        revoke_other_sessions: Optional[bool] = None,
        custom_claims: Optional[dict] = None,
        template_options: Optional[
            dict
        ] = None,  # for providing messaging template options (templates that are being sent via email / text message)
        template_id: Optional[
            str
        ] = None,  # for overriding the default template (templates that are being sent via email / text message)
    ):
        self.revokeOtherSessions = revoke_other_sessions
        self.customClaims = custom_claims
        self.templateOptions = template_options
        self.templateId = template_id


def signup_options_to_dict(signup_options: Optional[SignUpOptions] = None) -> dict:
    res: dict[str, Any] = {}
    if signup_options is not None:
        if signup_options.customClaims is not None:
            res["customClaims"] = signup_options.customClaims
        if signup_options.templateId is not None:
            res["templateId"] = signup_options.templateId
        if signup_options.templateOptions is not None:
            res["templateOptions"] = signup_options.templateOptions
        if signup_options.revokeOtherSessions is not None:
            res["revokeOtherSessions"] = signup_options.revokeOtherSessions
    return res


def get_method_string(method: DeliveryMethod) -> str:
    name = {
        DeliveryMethod.EMAIL: "email",
        DeliveryMethod.SMS: "sms",
        DeliveryMethod.VOICE: "voice",
        DeliveryMethod.WHATSAPP: "whatsapp",
        DeliveryMethod.EMBEDDED: "Embedded",
    }.get(method)

    if not name:
        raise AuthException(
            400, ERROR_TYPE_INVALID_ARGUMENT, f"Unknown delivery method: {method}"
        )

    return name



from typing import Any, List, Optional

from descope._http_base import HTTPBase
from descope.management.common import MgmtV1, SessionExpirationUnit, TenantAuthType


class Tenant(HTTPBase):
    def create(
        self,
        name: str,
        id: Optional[str] = None,
        self_provisioning_domains: Optional[List[str]] = None,
        custom_attributes: Optional[dict] = None,
        enforce_sso: Optional[bool] = False,
        disabled: Optional[bool] = False,
    ) -> dict:
        """
        Create a new tenant with the given name. Tenant IDs are provisioned automatically, but can be provided
        explicitly if needed. Both the name and ID must be unique per project.

        Args:
        name (str): The tenant's name
        id (str): Optional tenant ID.
        self_provisioning_domains (List[str]): An optional list of domain that are associated with this tenant.
            Users authenticating from these domains will be associated with this tenant.
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        enforce_sso (bool): Optional, login to the tenant is possible only using the configured sso
        disabled (bool): Optional, login to the tenant will be disabled

        Return value (dict):
        Return dict in the format
             {"id": <id>}

        Raise:
        AuthException: raised if creation operation fails
        """
        self_provisioning_domains = (
            [] if self_provisioning_domains is None else self_provisioning_domains
        )

        response = self._http.post(
            MgmtV1.tenant_create_path,
            body=Tenant._compose_create_update_body(
                name,
                id,
                self_provisioning_domains,
                custom_attributes,
                enforce_sso,
                disabled,
            ),
        )
        return response.json()

    def update(
        self,
        id: str,
        name: str,
        self_provisioning_domains: Optional[List[str]] = None,
        custom_attributes: Optional[dict] = None,
        enforce_sso: Optional[bool] = False,
        disabled: Optional[bool] = False,
    ):
        """
        Update an existing tenant with the given name and domains. IMPORTANT: All parameters are used as overrides
        to the existing tenant. Empty fields will override populated fields. Use carefully.

        Args:
        id (str): The ID of the tenant to update.
        name (str): Updated tenant name
        self_provisioning_domains (List[str]): An optional list of domain that are associated with this tenant.
            Users authenticating from these domains will be associated with this tenant.
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        enforce_sso (bool): Optional, login to the tenant is possible only using the configured sso
        disabled (bool): Optional, login to the tenant will be disabled

        Raise:
        AuthException: raised if creation operation fails
        """
        self_provisioning_domains = (
            [] if self_provisioning_domains is None else self_provisioning_domains
        )

        self._http.post(
            MgmtV1.tenant_update_path,
            body=Tenant._compose_create_update_body(
                name,
                id,
                self_provisioning_domains,
                custom_attributes,
                enforce_sso,
                disabled,
            ),
        )

    def update_settings(
        self,
        id: str,
        self_provisioning_domains: List[str],
        domains: Optional[List[str]] = None,
        auth_type: Optional[TenantAuthType] = None,
        session_settings_enabled: Optional[bool] = None,
        refresh_token_expiration: Optional[int] = None,
        refresh_token_expiration_unit: Optional[SessionExpirationUnit] = None,
        session_token_expiration: Optional[int] = None,
        session_token_expiration_unit: Optional[SessionExpirationUnit] = None,
        stepup_token_expiration: Optional[int] = None,
        stepup_token_expiration_unit: Optional[SessionExpirationUnit] = None,
        enable_inactivity: Optional[bool] = None,
        inactivity_time: Optional[int] = None,
        inactivity_time_unit: Optional[SessionExpirationUnit] = None,
        JITDisabled: Optional[bool] = None,
    ):
        """
        Update an existing tenant's session settings.

        Args:
            id (str): The ID of the tenant to update.
            self_provisioning_domains (List[str]): Domains for self-provisioning.
            domains (Optional[List[str]]): List of domains associated with the tenant.
            auth_type (Optional[TenantAuthType]): Authentication type for the tenant.
            session_settings_enabled (Optional[bool]): Whether session settings are enabled.
            refresh_token_expiration (Optional[int]): Expiration time for refresh tokens.
            refresh_token_expiration_unit (Optional[SessionExiprationUnit]): Unit for refresh token expiration.
            session_token_expiration (Optional[int]): Expiration time for session tokens.
            session_token_expiration_unit (Optional[SessionExiprationUnit]): Unit for session token expiration.
            stepup_token_expiration (Optional[int]): Expiration time for step-up tokens.
            stepup_token_expiration_unit (Optional[SessionExiprationUnit]): Unit for step-up token expiration.
            enable_inactivity (Optional[bool]): Whether inactivity timeout is enabled.
            inactivity_time (Optional[int]): Inactivity timeout duration.
            inactivity_time_unit (Optional[SessionExiprationUnit]): Unit for inactivity timeout.
            JITDisabled (Optional[bool]): Whether JIT is disabled.

        Raise:
            AuthException: raised if update operation fails
        """
        body: dict[str, Any] = {
            "tenantId": id,
            "selfProvisioningDomains": self_provisioning_domains,
            "domains": domains,
            "authType": auth_type,
            "enabled": session_settings_enabled,
            "refreshTokenExpiration": refresh_token_expiration,
            "refreshTokenExpirationUnit": refresh_token_expiration_unit,
            "sessionTokenExpiration": session_token_expiration,
            "sessionTokenExpirationUnit": session_token_expiration_unit,
            "stepupTokenExpiration": stepup_token_expiration,
            "stepupTokenExpirationUnit": stepup_token_expiration_unit,
            "enableInactivity": enable_inactivity,
            "inactivityTime": inactivity_time,
            "inactivityTimeUnit": inactivity_time_unit,
            "JITDisabled": JITDisabled,
        }

        body = {k: v for k, v in body.items() if v is not None}

        self._http.post(MgmtV1.tenant_settings_path, body=body, params=None)

    def delete(
        self,
        id: str,
        cascade: bool = False,
    ):
        """
        Delete an existing tenant. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        id (str): The ID of the tenant that's to be deleted.

        Raise:
        AuthException: raised if creation operation fails
        """
        self._http.post(
            MgmtV1.tenant_delete_path,
            body={"id": id, "cascade": cascade},
        )

    def load(
        self,
        id: str,
    ) -> dict:
        """
        Load tenant by id.

        Args:
        id (str): The ID of the tenant to load.

        Return value (dict):
        Return dict in the format
             {"id": <id>, "name": <name>, "selfProvisioningDomains": [], "customAttributes: {}, "createdTime": <timestamp>}
        Containing the loaded tenant information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(
            MgmtV1.tenant_load_path,
            params={"id": id},
        )
        return response.json()

    def load_settings(
        self,
        id: str,
    ) -> dict:
        """
        Load tenant session settings by id.

        Args:
        id (str): The ID of the tenant to load session settings for.

        Return value (dict):
        Return dict in the format
            { "domains":<list[str]>, "selfProvisioningDomains":<list[str]>, "authType":<str>,
             "enabled":<bool>, "refreshTokenExpiration":<int>, "refreshTokenExpirationUnit":<str>,
             "sessionTokenExpiration":<int>, "sessionTokenExpirationUnit":<str>,
             "stepupTokenExpiration":<int>, "stepupTokenExpirationUnit":<str>,
             "enableInactivity":<bool>, "inactivityTime":<int>, "inactivityTimeUnit":<str>,
             "JITDisabled":<bool> }
        Containing the loaded tenant session settings.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(
            MgmtV1.tenant_settings_path,
            params={"id": id},
        )
        return response.json()

    def load_all(
        self,
    ) -> dict:
        """
        Load all tenants.

        Return value (dict):
        Return dict in the format
             {"tenants": [{"id": <id>, "name": <name>, "selfProvisioningDomains": [], customAttributes: {}, "createdTime": <timestamp>}]}
        Containing the loaded tenant information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(
            MgmtV1.tenant_load_all_path,
        )
        return response.json()

    def search_all(
        self,
        ids: Optional[List[str]] = None,
        names: Optional[List[str]] = None,
        self_provisioning_domains: Optional[List[str]] = None,
        custom_attributes: Optional[dict] = None,
    ) -> dict:
        """
        Search all tenants.

        Args:
        ids (List[str]): Optional list of tenant IDs to filter by
        names (List[str]): Optional list of names to filter by
        self_provisioning_domains (List[str]): Optional list of self provisioning domains to filter by
        custom_attributes (dict): Optional search for a attribute with a given value

        Return value (dict):
        Return dict in the format
             {"tenants": [{"id": <id>, "name": <name>, "selfProvisioningDomains": [], customAttributes:{}}]}
        Containing the loaded tenant information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.post(
            MgmtV1.tenant_search_all_path,
            body={
                "tenantIds": ids,
                "tenantNames": names,
                "tenantSelfProvisioningDomains": self_provisioning_domains,
                "customAttributes": custom_attributes,
            },
        )
        return response.json()

    @staticmethod
    def _compose_create_update_body(
        name: str,
        id: Optional[str],
        self_provisioning_domains: List[str],
        custom_attributes: Optional[dict] = None,
        enforce_sso: Optional[bool] = False,
        disabled: Optional[bool] = False,
    ) -> dict:
        body: dict[str, Any] = {
            "name": name,
            "id": id,
            "selfProvisioningDomains": self_provisioning_domains,
            "enforceSSO": enforce_sso,
            "disabled": disabled,
        }
        if custom_attributes is not None:
            body["customAttributes"] = custom_attributes
        return body



from __future__ import annotations

from typing import Callable, Iterable, Optional

import jwt

from descope.common import (
    COOKIE_DATA_NAME,
    REFRESH_SESSION_TOKEN_NAME,
    SESSION_TOKEN_NAME,
)


def adjust_properties(jwt_response: dict, user_jwt: bool) -> dict:
    """Normalize top-level fields on a JWT response.

    Copies permissions/roles/tenants from present token claims (session or refresh)
    to the top-level and sets projectId and userId/keyId for convenience.
    """
    # Save permissions, roles and tenants info from Session token or from refresh token on the json top level
    if SESSION_TOKEN_NAME in jwt_response:
        jwt_response["permissions"] = jwt_response[SESSION_TOKEN_NAME].get(
            "permissions", []
        )
        jwt_response["roles"] = jwt_response[SESSION_TOKEN_NAME].get("roles", [])
        jwt_response["tenants"] = jwt_response[SESSION_TOKEN_NAME].get("tenants", {})
    elif REFRESH_SESSION_TOKEN_NAME in jwt_response:
        jwt_response["permissions"] = jwt_response[REFRESH_SESSION_TOKEN_NAME].get(
            "permissions", []
        )
        jwt_response["roles"] = jwt_response[REFRESH_SESSION_TOKEN_NAME].get(
            "roles", []
        )
        jwt_response["tenants"] = jwt_response[REFRESH_SESSION_TOKEN_NAME].get(
            "tenants", {}
        )
    else:
        jwt_response["permissions"] = jwt_response.get("permissions", [])
        jwt_response["roles"] = jwt_response.get("roles", [])
        jwt_response["tenants"] = jwt_response.get("tenants", {})

    # Save the projectID also in the dict top level
    issuer = (
        jwt_response.get(SESSION_TOKEN_NAME, {}).get("iss", None)
        or jwt_response.get(REFRESH_SESSION_TOKEN_NAME, {}).get("iss", None)
        or jwt_response.get("iss", "")
    )
    jwt_response["projectId"] = issuer.rsplit("/")[
        -1
    ]  # support both url issuer and project ID issuer

    sub = (
        jwt_response.get(SESSION_TOKEN_NAME, {}).get("dsub", None)
        or jwt_response.get(SESSION_TOKEN_NAME, {}).get("sub", None)
        or jwt_response.get(REFRESH_SESSION_TOKEN_NAME, {}).get("dsub", None)
        or jwt_response.get(REFRESH_SESSION_TOKEN_NAME, {}).get("sub", None)
        or jwt_response.get("sub", "")
    )
    if user_jwt:
        # Save the userID also in the dict top level
        jwt_response["userId"] = sub
    else:
        # Save the AccessKeyID also in the dict top level
        jwt_response["keyId"] = sub

    return jwt_response


def decode_token_unverified(
    token: str, audience: Optional[str | Iterable[str]] = None
) -> dict:
    """Decode a JWT without verifying signature (used when no validator is provided).

    Audience verification is disabled by default since no key is provided.
    Returns an empty dict if decoding fails.
    """
    try:
        return jwt.decode(
            token, options={"verify_signature": False, "verify_aud": False}
        )
    except Exception:
        return {}


def generate_auth_info(
    response_body: dict,
    refresh_token: Optional[str],
    user_jwt: bool,
    audience: Optional[str | Iterable[str]] = None,
    token_validator: Optional[
        Callable[[str, Optional[str | Iterable[str]]], dict]
    ] = None,
) -> dict:
    """Build the normalized JWT info dict using a provided token validator.

    token_validator should accept (token, audience) and return decoded claims dict.
    If not provided, tokens will be decoded without signature verification.
    """
    if token_validator is None:
        token_validator = decode_token_unverified

    jwt_response: dict = {}

    st_jwt = response_body.get("sessionJwt", "")
    if st_jwt:
        jwt_response[SESSION_TOKEN_NAME] = token_validator(st_jwt, audience)

    rt_jwt = response_body.get("refreshJwt", "")
    if rt_jwt:
        jwt_response[REFRESH_SESSION_TOKEN_NAME] = token_validator(rt_jwt, audience)
    elif refresh_token:
        jwt_response[REFRESH_SESSION_TOKEN_NAME] = token_validator(
            refresh_token, audience
        )

    jwt_response = adjust_properties(jwt_response, user_jwt)

    if user_jwt:
        jwt_response[COOKIE_DATA_NAME] = {
            "exp": response_body.get("cookieExpiration", 0),
            "maxAge": response_body.get("cookieMaxAge", 0),
            "domain": response_body.get("cookieDomain", ""),
            "path": response_body.get("cookiePath", "/"),
        }

    return jwt_response


def generate_jwt_response(
    response_body: dict,
    refresh_cookie: Optional[str],
    audience: Optional[str | Iterable[str]] = None,
    token_validator: Optional[
        Callable[[str, Optional[str | Iterable[str]]], dict]
    ] = None,
) -> dict:
    """Compose the final JWT response body using the provided token validator."""
    jwt_response = generate_auth_info(
        response_body, refresh_cookie, True, audience, token_validator
    )
    jwt_response["user"] = response_body.get("user", {})
    jwt_response["firstSeen"] = response_body.get("firstSeen", True)
    return jwt_response



from __future__ import annotations

from typing import List, Optional

from descope._http_base import HTTPBase
from descope.management.common import MgmtV1


class Project(HTTPBase):
    def update_name(
        self,
        name: str,
    ):
        """
        Update the current project name.

        Args:
        name (str):  The new name for the project.
        Raise:
        AuthException: raised if operation fails
        """
        self._http.post(
            MgmtV1.project_update_name,
            body={
                "name": name,
            },
        )

    def update_tags(
        self,
        tags: List[str],
    ):
        """
        Update the current project tags.

        Args:
        tags (List[str]):  Array of free text tags.
        Raise:
        AuthException: raised if operation fails
        """
        self._http.post(
            MgmtV1.project_update_tags,
            body={
                "tags": tags,
            },
        )

    def list_projects(
        self,
    ) -> dict:
        """
        List of all the projects in the company.

        Return value (dict):
        Return dict in the format
             {"projects": []}
        "projects" contains a list of all of the projects and their information

        Raise:
        AuthException: raised if operation fails
        """
        response = self._http.post(
            MgmtV1.project_list_projects,
            body={},
        )
        resp = response.json()

        projects = resp["projects"]
        # Apply the function to the projects list
        formatted_projects = self.remove_tag_field(projects)

        # Return the same structure with 'tag' removed
        result = {"projects": formatted_projects}
        return result

    def clone(
        self,
        name: str,
        environment: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ):
        """
        Clone the current project, including its settings and configurations.
        - This action is supported only with a pro license or above.
        - Users, tenants and access keys are not cloned.

        Args:
        name (str): The new name for the project.
        environment (str): Optional state for the project. Currently, only the "production" tag is supported.
        tags(list[str]): Optional free text tags.

        Return value (dict):
        Return dict Containing the new project details (name, id, environment and tag).

        Raise:
        AuthException: raised if clone operation fails
        """
        response = self._http.post(
            MgmtV1.project_clone,
            body={
                "name": name,
                "environment": environment,
                "tags": tags,
            },
        )
        return response.json()

    def export_project(
        self,
    ):
        """
        Exports all settings and configurations for a project and returns the
        raw JSON files response as a dictionary.
        - This action is supported only with a pro license or above.
        - Users, tenants and access keys are not cloned.
        - Secrets, keys and tokens are not stripped from the exported data.

        Return value (dict):
        Return dict Containing the exported JSON files payload.

        Raise:
        AuthException: raised if export operation fails
        """
        response = self._http.post(
            MgmtV1.project_export,
            body={},
        )
        return response.json()["files"]

    def import_project(
        self,
        files: dict,
    ):
        """
        Imports all settings and configurations for a project overriding any current
        configuration.
        - This action is supported only with a pro license or above.
        - Secrets, keys and tokens are not overwritten unless overwritten in the input.

        Args:
        files (dict): The raw JSON dictionary of files, in the same format as the one
        returned by calls to export.

        Raise:
        AuthException: raised if import operation fails
        """
        self._http.post(
            MgmtV1.project_import,
            body={
                "files": files,
            },
        )
        return

    # Function to remove 'tag' field from each project
    def remove_tag_field(self, projects):
        return [
            {k: v for k, v in project.items() if k != "tag"} for project in projects
        ]



from __future__ import annotations

import os
import platform
from http import HTTPStatus
from typing import Optional, Union, cast

try:
    from importlib.metadata import version
except ImportError:  # pragma: no cover
    from pkg_resources import get_distribution

import requests

from descope.common import (
    DEFAULT_BASE_URL,
    DEFAULT_DOMAIN,
    DEFAULT_TIMEOUT_SECONDS,
    DEFAULT_URL_PREFIX,
)
from descope.exceptions import (
    API_RATE_LIMIT_RETRY_AFTER_HEADER,
    ERROR_TYPE_API_RATE_LIMIT,
    ERROR_TYPE_INVALID_ARGUMENT,
    ERROR_TYPE_SERVER_ERROR,
    AuthException,
    RateLimitException,
)


def sdk_version():
    try:
        return version("descope")  # type: ignore
    except NameError:  # pragma: no cover
        return get_distribution("descope").version  # type: ignore


_default_headers = {
    "Content-Type": "application/json",
    "x-descope-sdk-name": "python",
    "x-descope-sdk-python-version": platform.python_version(),
    "x-descope-sdk-version": sdk_version(),
}


class HTTPClient:
    def __init__(
        self,
        project_id: str,
        base_url: Optional[str] = None,
        *,
        timeout_seconds: float = DEFAULT_TIMEOUT_SECONDS,
        secure: bool = True,
        management_key: Optional[str] = None,
    ) -> None:
        if not project_id:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                (
                    "Project ID is required to initialize HTTP client"
                    "Set environment variable DESCOPE_PROJECT_ID or pass your Project ID to the init function."
                ),
            )

        # Prefer explicitly provided base_url, then env var, then computed default
        env_base = os.getenv("DESCOPE_BASE_URI")
        self.base_url = base_url or env_base or self.base_url_for_project_id(project_id)

        self.project_id = project_id
        self.timeout_seconds = timeout_seconds
        self.secure = secure
        self.management_key = management_key

    # ------------- public API -------------
    def get(
        self,
        uri: str,
        *,
        params=None,
        allow_redirects: Optional[bool] = True,
        pswd: Optional[str] = None,
    ) -> requests.Response:
        response = requests.get(
            f"{self.base_url}{uri}",
            headers=self._get_default_headers(pswd),
            params=params,
            allow_redirects=cast(bool, allow_redirects),
            verify=self.secure,
            timeout=self.timeout_seconds,
        )
        self._raise_from_response(response)
        return response

    def post(
        self,
        uri: str,
        *,
        body: Optional[Union[dict, list[dict], list[str]]] = None,
        params=None,
        pswd: Optional[str] = None,
        base_url: Optional[str] = None,
    ) -> requests.Response:
        response = requests.post(
            f"{base_url or self.base_url}{uri}",
            headers=self._get_default_headers(pswd),
            json=body,
            allow_redirects=False,
            verify=self.secure,
            params=params,
            timeout=self.timeout_seconds,
        )
        self._raise_from_response(response)
        return response

    def patch(
        self,
        uri: str,
        *,
        body: Optional[Union[dict, list[dict], list[str]]],
        params=None,
        pswd: Optional[str] = None,
    ) -> requests.Response:
        response = requests.patch(
            f"{self.base_url}{uri}",
            headers=self._get_default_headers(pswd),
            json=body,
            allow_redirects=False,
            verify=self.secure,
            params=params,
            timeout=self.timeout_seconds,
        )
        self._raise_from_response(response)
        return response

    def delete(
        self,
        uri: str,
        *,
        params=None,
        pswd: Optional[str] = None,
    ) -> requests.Response:
        response = requests.delete(
            f"{self.base_url}{uri}",
            params=params,
            headers=self._get_default_headers(pswd),
            allow_redirects=False,
            verify=self.secure,
            timeout=self.timeout_seconds,
        )
        self._raise_from_response(response)
        return response

    def get_default_headers(self, pswd: Optional[str] = None) -> dict:
        return self._get_default_headers(pswd)

    # ------------- helpers -------------
    @staticmethod
    def base_url_for_project_id(project_id: str) -> str:
        if len(project_id) >= 32:
            region = project_id[1:5]
            return ".".join([DEFAULT_URL_PREFIX, region, DEFAULT_DOMAIN])
        return DEFAULT_BASE_URL

    def _parse_retry_after(self, headers) -> int:
        try:
            return int(headers.get(API_RATE_LIMIT_RETRY_AFTER_HEADER, 0))
        except (ValueError, TypeError):
            return 0

    def _raise_rate_limit_exception(self, response):
        try:
            resp = response.json()
            raise RateLimitException(
                resp.get("errorCode", HTTPStatus.TOO_MANY_REQUESTS),
                ERROR_TYPE_API_RATE_LIMIT,
                resp.get("errorDescription", ""),
                resp.get("errorMessage", ""),
                rate_limit_parameters={
                    API_RATE_LIMIT_RETRY_AFTER_HEADER: self._parse_retry_after(
                        response.headers
                    )
                },
            )
        except RateLimitException:
            raise
        except Exception:
            raise RateLimitException(
                status_code=HTTPStatus.TOO_MANY_REQUESTS,
                error_type=ERROR_TYPE_API_RATE_LIMIT,
                error_message=ERROR_TYPE_API_RATE_LIMIT,
                error_description=ERROR_TYPE_API_RATE_LIMIT,
            )

    def _raise_from_response(self, response):
        if response.ok:
            return
        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            self._raise_rate_limit_exception(response)
        raise AuthException(
            response.status_code,
            ERROR_TYPE_SERVER_ERROR,
            response.text,
        )

    def _get_default_headers(self, pswd: Optional[str] = None):
        headers = _default_headers.copy()
        headers["x-descope-project-id"] = self.project_id
        bearer = self.project_id
        if pswd:
            bearer = f"{self.project_id}:{pswd}"
        if self.management_key:
            bearer = f"{bearer}:{self.management_key}"
        headers["Authorization"] = f"Bearer {bearer}"
        return headers



from __future__ import annotations

from descope.http_client import HTTPClient


class HTTPBase:
    """Base class for classes that only need HTTP access."""

    def __init__(self, http_client: HTTPClient):
        self._http = http_client



from typing import List, Optional

from descope._http_base import HTTPBase
from descope.management.common import MgmtV1


class RoleMapping:
    """Map IDP group names to the Descope role name"""

    def __init__(self, groups: List[str], role_name: str):
        self.groups = groups
        self.role_name = role_name


class AttributeMapping:
    """Map Descope user attributes to IDP user attributes"""

    def __init__(
        self,
        name: Optional[str] = None,
        email: Optional[str] = None,
        phone_number: Optional[str] = None,
        group: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        picture: Optional[str] = None,
        custom_attributes: Optional[dict] = None,
    ):
        self.name = name
        self.email = email
        self.phone_number = phone_number
        self.group = group
        self.given_name = given_name
        self.middle_name = middle_name
        self.family_name = family_name
        self.picture = picture
        self.custom_attributes = custom_attributes


class OIDCAttributeMapping:
    """
    Represents tenant OIDC attribute mapping.
    """

    def __init__(
        self,
        login_id: str,
        name: str,
        given_name: str,
        middle_name: str,
        family_name: str,
        email: str,
        verified_email: str,
        username: str,
        phone_number: str,
        verified_phone: str,
        picture: str,
    ):
        self.login_id = login_id
        self.name = name
        self.given_name = given_name
        self.middle_name = middle_name
        self.family_name = family_name
        self.email = email
        self.verified_email = verified_email
        self.username = username
        self.phone_number = phone_number
        self.verified_phone = verified_phone
        self.picture = picture


class SSOOIDCSettings:
    """
    Represents tenant OIDC settings.
    """

    def __init__(
        self,
        name: str,
        client_id: str,
        client_secret: Optional[str] = None,
        redirect_url: Optional[str] = None,
        auth_url: Optional[str] = None,
        token_url: Optional[str] = None,
        user_data_url: Optional[str] = None,
        scope: Optional[List[str]] = None,
        jwks_url: Optional[str] = None,
        attribute_mapping: Optional[OIDCAttributeMapping] = None,
        manage_provider_tokens: Optional[bool] = False,
        callback_domain: Optional[str] = None,
        prompt: Optional[List[str]] = None,
        grant_type: Optional[str] = None,
        issuer: Optional[str] = None,
    ):
        self.name = name
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_url = redirect_url
        self.auth_url = auth_url
        self.token_url = token_url
        self.user_data_url = user_data_url
        self.scope = scope
        self.jwks_url = jwks_url
        self.attribute_mapping = attribute_mapping
        self.manage_provider_tokens = manage_provider_tokens
        self.callback_domain = callback_domain
        self.prompt = prompt
        self.grant_type = grant_type
        self.issuer = issuer


class SSOSAMLSettings:
    """
    Represents tenant SAML settings (manually configuration).
    """

    def __init__(
        self,
        idp_url: str,
        idp_entity_id: str,
        idp_cert: str,
        attribute_mapping: Optional[AttributeMapping] = None,
        role_mappings: Optional[List[RoleMapping]] = None,
        default_sso_roles: Optional[List[str]] = None,
        # NOTICE - the following fields should be overridden only in case of SSO migration, otherwise, do not modify these fields
        sp_acs_url: Optional[str] = None,
        sp_entity_id: Optional[str] = None,
    ):
        self.idp_url = idp_url
        self.idp_entity_id = idp_entity_id
        self.idp_cert = idp_cert
        self.attribute_mapping = attribute_mapping
        self.role_mappings = role_mappings
        self.default_sso_roles = default_sso_roles
        self.sp_acs_url = sp_acs_url
        self.sp_entity_id = sp_entity_id


class SSOSAMLSettingsByMetadata:
    """
    Represents tenant SAML settings (automatically (by metadata xml) configuration).
    """

    def __init__(
        self,
        idp_metadata_url: str,
        attribute_mapping: Optional[AttributeMapping] = None,
        role_mappings: Optional[List[RoleMapping]] = None,
        default_sso_roles: Optional[List[str]] = None,
        # NOTICE - the following fields should be overridden only in case of SSO migration, otherwise, do not modify these fields
        sp_acs_url: Optional[str] = None,
        sp_entity_id: Optional[str] = None,
    ):
        self.idp_metadata_url = idp_metadata_url
        self.attribute_mapping = attribute_mapping
        self.role_mappings = role_mappings
        self.default_sso_roles = default_sso_roles
        self.sp_acs_url = sp_acs_url
        self.sp_entity_id = sp_entity_id


class SSOSettings(HTTPBase):
    def load_settings(
        self,
        tenant_id: str,
    ) -> dict:
        """
        Load SSO setting for the provided tenant_id.

        Args:
        tenant_id (str): The tenant ID of the desired SSO Settings

        Return value (dict):
        Containing the loaded SSO settings information.
        Return dict in the format:
             {"tenant": {"id": "T2AAAA", "name": "myTenantName", "selfProvisioningDomains": [], "customAttributes": {}, "authType": "saml", "domains": ["lulu", "kuku"]}, "saml": {"idpEntityId": "", "idpSSOUrl": "", "idpCertificate": "", "idpMetadataUrl": "https://dummy.com/metadata", "spEntityId": "", "spACSUrl": "", "spCertificate": "", "attributeMapping": {"name": "name", "email": "email", "username": "", "phoneNumber": "phone", "group": "", "givenName": "", "middleName": "", "familyName": "", "picture": "", "customAttributes": {}}, "groupsMapping": [], "redirectUrl": ""}, "oidc": {"name": "", "clientId": "", "clientSecret": "", "redirectUrl": "", "authUrl": "", "tokenUrl": "", "userDataUrl": "", "scope": [], "JWKsUrl": "", "userAttrMapping": {"loginId": "sub", "username": "", "name": "name", "email": "email", "phoneNumber": "phone_number", "verifiedEmail": "email_verified", "verifiedPhone": "phone_number_verified", "picture": "picture", "givenName": "given_name", "middleName": "middle_name", "familyName": "family_name"}, "manageProviderTokens": False, "callbackDomain": "", "prompt": [], "grantType": "authorization_code", "issuer": ""}}

        Raise:
        AuthException: raised if load configuration operation fails
        """
        response = self._http.get(
            uri=MgmtV1.sso_load_settings_path,
            params={"tenantId": tenant_id},
        )
        return response.json()

    def delete_settings(
        self,
        tenant_id: str,
    ):
        """
        Delete SSO setting for the provided tenant_id.

        Args:
        tenant_id (str): The tenant ID of the desired SSO Settings to delete

        Raise:
        AuthException: raised if delete operation fails
        """
        self._http.delete(
            MgmtV1.sso_settings_path,
            params={"tenantId": tenant_id},
        )

    def configure_oidc_settings(
        self,
        tenant_id: str,
        settings: SSOOIDCSettings,
        domains: Optional[List[str]] = None,
    ):
        """
        Configure SSO OIDC settings for a tenant.

        Args:
        tenant_id (str): The tenant ID to be configured
        settings (SSOOIDCSettings): The OIDC settings to be configured for this tenant (all settings parameters are required).
        domains (List[str]): Optional,domains used to associate users authenticating via SSO with this tenant. Use empty list or None to reset them.

        Raise:
        AuthException: raised if configuration operation fails
        """

        self._http.post(
            MgmtV1.sso_configure_oidc_settings,
            body=SSOSettings._compose_configure_oidc_settings_body(
                tenant_id, settings, domains
            ),
        )

    def configure_saml_settings(
        self,
        tenant_id: str,
        settings: SSOSAMLSettings,
        redirect_url: Optional[str] = None,
        domains: Optional[List[str]] = None,
    ):
        """
        Configure SSO SAML settings for a tenant.

        Args:
        tenant_id (str): The tenant ID to be configured
        settings (SSOSAMLSettings): The SAML settings to be configured for this tenant (all settings parameters are required).
        redirect_url (str): Optional,the Redirect URL to use after successful authentication, or empty string to reset it (if not given it has to be set when starting an SSO authentication via the request).
        domains (List[str]): Optional, domains used to associate users authenticating via SSO with this tenant. Use empty list or None to reset them.

        Raise:
        AuthException: raised if configuration operation fails
        """

        self._http.post(
            MgmtV1.sso_configure_saml_settings,
            body=SSOSettings._compose_configure_saml_settings_body(
                tenant_id, settings, redirect_url, domains
            ),
        )

    def configure_saml_settings_by_metadata(
        self,
        tenant_id: str,
        settings: SSOSAMLSettingsByMetadata,
        redirect_url: Optional[str] = None,
        domains: Optional[List[str]] = None,
    ):
        """
        Configure SSO SAML settings for a tenant by fetching them from an IDP metadata URL.

        Args:
        tenant_id (str): The tenant ID to be configured
        settings (SSOSAMLSettingsByMetadata): The SAML settings to be configured for this tenant (all settings parameters are required).
        redirect_url (str): Optional, the Redirect URL to use after successful authentication, or empty string to reset it (if not given it has to be set when starting an SSO authentication via the request).
        domains (List[str]): Optional, domains used to associate users authenticating via SSO with this tenant. Use empty list or None to reset them.

        Raise:
        AuthException: raised if configuration operation fails
        """

        self._http.post(
            MgmtV1.sso_configure_saml_by_metadata_settings,
            body=SSOSettings._compose_configure_saml_settings_by_metadata_body(
                tenant_id, settings, redirect_url, domains
            ),
        )

    # DEPRECATED
    def get_settings(
        self,
        tenant_id: str,
    ) -> dict:
        """
        DEPRECATED (use load_settings(..) function instead)

        Get SSO setting for the provided tenant_id.

        Args:
        tenant_id (str): The tenant ID of the desired SSO Settings

        Return value (dict):
        Containing the loaded SSO settings information.

        Raise:
        AuthException: raised if configuration operation fails
        """
        response = self._http.get(
            uri=MgmtV1.sso_settings_path,
            params={"tenantId": tenant_id},
        )
        return response.json()

    # DEPRECATED
    def configure(
        self,
        tenant_id: str,
        idp_url: str,
        entity_id: str,
        idp_cert: str,
        redirect_url: str,
        domains: Optional[List[str]] = None,
    ) -> None:
        """
        DEPRECATED (use configure_saml_settings(..) function instead)

        Configure SSO setting for a tenant manually. Alternatively, `configure_via_metadata` can be used instead.

        Args:
        tenant_id (str): The tenant ID to be configured
        idp_url (str): The URL for the identity provider.
        entity_id (str): The entity ID (in the IDP).
        idp_cert (str): The certificate provided by the IDP.
        redirect_url (str): The Redirect URL to use after successful authentication, or empty string to reset it.
        domain (List[str]): domains used to associate users authenticating via SSO with this tenant. Use empty list or None to reset them.

        Raise:
        AuthException: raised if configuration operation fails
        """
        self._http.post(
            MgmtV1.sso_settings_path,
            body=SSOSettings._compose_configure_body(
                tenant_id, idp_url, entity_id, idp_cert, redirect_url, domains
            ),
        )

    # DEPRECATED
    def configure_via_metadata(
        self,
        tenant_id: str,
        idp_metadata_url: str,
        redirect_url: Optional[str] = None,
        domains: Optional[List[str]] = None,
    ):
        """
        DEPRECATED (use configure_saml_settings_by_metadata(..) function instead)

        Configure SSO setting for am IDP metadata URL. Alternatively, `configure` can be used instead.

        Args:
        tenant_id (str): The tenant ID to be configured
        idp_metadata_url (str): The URL to fetch SSO settings from.
        redirect_url (str): The Redirect URL to use after successful authentication, or empty string to reset it.
        domains (List[str]): domains used to associate users authenticating via SSO with this tenant. Use empty list or None to reset them.

        Raise:
        AuthException: raised if configuration operation fails
        """
        self._http.post(
            MgmtV1.sso_metadata_path,
            body=SSOSettings._compose_metadata_body(
                tenant_id, idp_metadata_url, redirect_url, domains
            ),
        )

    # DEPRECATED
    def mapping(
        self,
        tenant_id: str,
        role_mappings: Optional[List[RoleMapping]] = None,
        attribute_mapping: Optional[AttributeMapping] = None,
    ):
        """
        DEPRECATED (use configure_saml_settings(..) or configure_saml_settings_by_metadata(..) functions instead)

        Configure SSO role mapping from the IDP groups to the Descope roles.

        Args:
        tenant_id (str): The tenant ID to be configured
        role_mappings (List[RoleMapping]): A mapping between IDP groups and Descope roles.
        attribute_mapping (AttributeMapping): A mapping between IDP user attributes and descope attributes.

        Raise:
        AuthException: raised if configuration operation fails
        """
        self._http.post(
            MgmtV1.sso_mapping_path,
            body=SSOSettings._compose_mapping_body(
                tenant_id, role_mappings, attribute_mapping
            ),
        )

    @staticmethod
    def _compose_configure_body(
        tenant_id: str,
        idp_url: str,
        entity_id: str,
        idp_cert: str,
        redirect_url: str,
        domains: Optional[List[str]],
    ) -> dict:
        return {
            "tenantId": tenant_id,
            "idpURL": idp_url,
            "entityId": entity_id,
            "idpCert": idp_cert,
            "redirectURL": redirect_url,
            "domains": domains,
        }

    @staticmethod
    def _compose_metadata_body(
        tenant_id: str,
        idp_metadata_url: str,
        redirect_url: Optional[str] = None,
        domains: Optional[List[str]] = None,
    ) -> dict:
        return {
            "tenantId": tenant_id,
            "idpMetadataURL": idp_metadata_url,
            "redirectURL": redirect_url,
            "domains": domains,
        }

    @staticmethod
    def _compose_mapping_body(
        tenant_id: str,
        role_mapping: Optional[List[RoleMapping]],
        attribute_mapping: Optional[AttributeMapping],
    ) -> dict:
        return {
            "tenantId": tenant_id,
            "roleMappings": SSOSettings._role_mapping_to_dict(role_mapping),
            "attributeMapping": SSOSettings._attribute_mapping_to_dict(
                attribute_mapping
            ),
        }

    @staticmethod
    def _role_mapping_to_dict(role_mapping: Optional[List[RoleMapping]]) -> list:
        if role_mapping is None:
            role_mapping = []
        role_mapping_list = []
        for mapping in role_mapping:
            role_mapping_list.append(
                {
                    "groups": mapping.groups,
                    "roleName": mapping.role_name,
                }
            )
        return role_mapping_list

    @staticmethod
    def _attribute_mapping_to_dict(
        attribute_mapping: Optional[AttributeMapping],
    ) -> dict:
        if attribute_mapping is None:
            raise ValueError("Attribute mapping cannot be None")
        return {
            "name": attribute_mapping.name,
            "email": attribute_mapping.email,
            "phoneNumber": attribute_mapping.phone_number,
            "group": attribute_mapping.group,
            "givenName": attribute_mapping.given_name,
            "middleName": attribute_mapping.middle_name,
            "familyName": attribute_mapping.family_name,
            "picture": attribute_mapping.picture,
            "customAttributes": attribute_mapping.custom_attributes,
        }

    @staticmethod
    def _compose_configure_oidc_settings_body(
        tenant_id: str,
        settings: SSOOIDCSettings,
        domains: Optional[List[str]],
    ) -> dict:
        attr_mapping = None
        if settings.attribute_mapping:
            attr_mapping = {
                "loginId": settings.attribute_mapping.login_id,
                "name": settings.attribute_mapping.name,
                "givenName": settings.attribute_mapping.given_name,
                "middleName": settings.attribute_mapping.middle_name,
                "familyName": settings.attribute_mapping.family_name,
                "email": settings.attribute_mapping.email,
                "verifiedEmail": settings.attribute_mapping.verified_email,
                "username": settings.attribute_mapping.username,
                "phoneNumber": settings.attribute_mapping.phone_number,
                "verifiedPhone": settings.attribute_mapping.verified_phone,
                "picture": settings.attribute_mapping.picture,
            }

        return {
            "tenantId": tenant_id,
            "settings": {
                "name": settings.name,
                "clientId": settings.client_id,
                "clientSecret": settings.client_secret,
                "redirectUrl": settings.redirect_url,
                "authUrl": settings.auth_url,
                "tokenUrl": settings.token_url,
                "userDataUrl": settings.user_data_url,
                "scope": settings.scope,
                "JWKsUrl": settings.jwks_url,
                "userAttrMapping": attr_mapping,
                "manageProviderTokens": settings.manage_provider_tokens,
                "callbackDomain": settings.callback_domain,
                "prompt": settings.prompt,
                "grantType": settings.grant_type,
                "issuer": settings.issuer,
            },
            "domains": domains,
        }

    @staticmethod
    def _compose_configure_saml_settings_body(
        tenant_id: str,
        settings: SSOSAMLSettings,
        redirect_url: Optional[str],
        domains: Optional[List[str]],
    ) -> dict:
        attr_mapping = None
        if settings.attribute_mapping:
            attr_mapping = SSOSettings._attribute_mapping_to_dict(
                settings.attribute_mapping
            )

        return {
            "tenantId": tenant_id,
            "settings": {
                "idpUrl": settings.idp_url,
                "entityId": settings.idp_entity_id,
                "idpCert": settings.idp_cert,
                "spACSUrl": settings.sp_acs_url,
                "spEntityId": settings.sp_entity_id,
                "attributeMapping": attr_mapping,
                "roleMappings": SSOSettings._role_mapping_to_dict(
                    settings.role_mappings
                ),
                "defaultSSORoles": settings.default_sso_roles,
            },
            "redirectUrl": redirect_url,
            "domains": domains,
        }

    @staticmethod
    def _compose_configure_saml_settings_by_metadata_body(
        tenant_id: str,
        settings: SSOSAMLSettingsByMetadata,
        redirect_url: Optional[str],
        domains: Optional[List[str]],
    ) -> dict:
        attr_mapping = None
        if settings.attribute_mapping:
            attr_mapping = SSOSettings._attribute_mapping_to_dict(
                settings.attribute_mapping
            )

        return {
            "tenantId": tenant_id,
            "settings": {
                "idpMetadataUrl": settings.idp_metadata_url,
                "spACSUrl": settings.sp_acs_url,
                "spEntityId": settings.sp_entity_id,
                "attributeMapping": attr_mapping,
                "roleMappings": SSOSettings._role_mapping_to_dict(
                    settings.role_mappings
                ),
                "defaultSSORoles": settings.default_sso_roles,
            },
            "redirectUrl": redirect_url,
            "domains": domains,
        }
